from sklearn.metrics import confusion_matrix, roc_curve, auc, balanced_accuracy_score
import matplotlib.pyplot as plt 
import numpy as np
import torch
import torch.nn.functional as F
from HPO.utils.time_series_augmentation_torch import  jitter, scaling, window_warp,crop, cutout
from HPO.utils.train_utils import collate_fn_padd, collate_fn_padd_x
import random
import copy
from torch.utils.data import Sampler
from sklearn.model_selection import StratifiedKFold
import pandas as pd

class StratifiedSampler(Sampler):
    """Stratified Sampling
    Provides equal representation of target classes in each batch
    """
    def __init__(self, class_vector, batch_size):
        """
        Arguments
        ---------
        class_vector : torch tensor
            a vector of class labels
        batch_size : integer
            batch_size
        """
        self.n_splits = int(class_vector.size(0) / batch_size)
        self.class_vector = class_vector

    def gen_sample_array(self):
        try:
            from sklearn.model_selection import StratifiedShuffleSplit
        except:
           pass
           #print('Need scikit-learn for this functionality')
        import numpy as np
        
        s = StratifiedShuffleSplit(n_splits=self.n_splits, test_size=0.5)
        X = th.randn(self.class_vector.size(0),2).numpy()
        y = self.class_vector.numpy()
        s.get_n_splits(X, y)

        train_index, test_index = next(s.split(X, y))
        return np.hstack([train_index, test_index])

    def __iter__(self):
        return iter(self.gen_sample_array())

    def __len__(self):
        return len(self.class_vector)

class StratifiedBatchSampler:
    """Stratified batch sampling
    Provides equal representation of target classes in each batch
    """
    def __init__(self, y, batch_size, shuffle=True):
        if torch.is_tensor(y):
            y = y.cpu().numpy()
        assert len(y.shape) == 1, 'label array must be 1D'
       #print("Number of samples: {} -- Batch_Size: {}".format(len(y),batch_size))
        n_batches = int(len(y) / batch_size)
        self.skf = StratifiedKFold(n_splits=n_batches, shuffle=shuffle)
        self.X = torch.randn(len(y),1).numpy()
        self.y = y
        self.shuffle = shuffle

    def __iter__(self):
        if self.shuffle:
            self.skf.random_state = torch.randint(0,int(1e8),size=()).item()
        for train_idx, test_idx in self.skf.split(self.X, self.y):
            yield test_idx

    def __len__(self):
        return len(self.y)

def augment(x, y = None, settings = None ,device = None):
  augs = [jitter, scaling, window_warp,crop]
  
  rate = 0.5
  if device == None:
    for func in augs:
      if random.random() < rate:
        x,y = func(x,y)
  else:
    if settings != None:
      rates = settings["rates"]
      hps = settings["hps"]
      for rate, hp,func in zip(rates, hps,augs):
        if random.random() < rate:
          x,y = func(x,y,hp, device = device)
    else:
      for func in augs:
        if random.random() < rate:
          x,y = func(x,y, device = device)
  return x  
def ce_loss(logits, targets, use_hard_labels=True, reduction='none'):
    """
    wrapper for cross entropy loss in pytorch.
    Args:
        logits: logit values, shape=[Batch size, # of classes]
        targets: integer or vector, shape=[Batch size] or [Batch size, # of classes]
        use_hard_labels: If True, targets have [Batch size] shape with int values. If False, the target is vector (default True)
    """
    if use_hard_labels:
        log_pred = F.log_softmax(logits, dim=-1)
        return F.nll_loss(log_pred, targets, reduction=reduction)
        # return F.cross_entropy(logits, targets, reduction=reduction) this is unstable
    else:
        assert logits.shape == targets.shape
        log_pred = F.log_softmax(logits, dim=-1)
        nll_loss = torch.sum(-targets * log_pred, dim=1)
        return nll_loss

def consistency_loss(logits_s, logits_w, name='ce', T=1.0, p_cutoff=0.0, use_hard_labels=True):
    assert name in ['ce', 'L2']
    logits_w = logits_w.detach()
    if name == 'L2':
        assert logits_w.size() == logits_s.size()
        return F.mse_loss(logits_s, logits_w, reduction='mean')

    elif name == 'L2_mask':
        pass

    elif name == 'ce':
        pseudo_label = torch.softmax(logits_w, dim=-1)
        max_probs, max_idx = torch.max(pseudo_label, dim=-1)
        mask = max_probs.ge(p_cutoff).float()
        select = max_probs.ge(p_cutoff).long()
        # strong_prob, strong_idx = torch.max(torch.softmax(logits_s, dim=-1), dim=-1)
        # strong_select = strong_prob.ge(p_cutoff).long()
        # select = select * strong_select * (strong_idx == max_idx)
        if use_hard_labels:
            masked_loss = ce_loss(logits_s, max_idx, use_hard_labels, reduction='none') * mask
        else:
            pseudo_label = torch.softmax(logits_w / T, dim=-1)
            masked_loss = ce_loss(logits_s, pseudo_label, use_hard_labels) * mask
        return masked_loss.mean()
    else:
        assert Exception('Not Implemented consistency_loss')
            


class Evaluator:
  def __init__(self,batch_size,n_classes,cuda_device = None, testloader = None):
    out_data = []
    self.cuda_device = cuda_device
    self.testloader = testloader
    self.batch_size = batch_size
    self.correct = [] #Binary Array of correct/incorrect for each sample
    self.n_correct = 0 #Number of correct values
    self.n_incorrect= 0 # Number of incorrect values
    self.n_classes = n_classes
    self.n_total = 0 #Total number of values
    self.confusion_matrix = np.zeros(shape = (n_classes,n_classes)) #Matrix of prediction vs true values

  def regression(self,model,index,hp):
    path = hp["PATH"]
    outputs= []
    outputs_raw= []
    raw_data = torch.swapaxes(self.testloader.dataset.x,1,0).detach().cpu().numpy()
    lossFn = F.mse_loss
    loss = 0
    labels = []
    with torch.no_grad():
      for sample, label in self.testloader:
        out_raw = model(sample)
        out = F.tanh(model(sample))
        labels.append(label)
        loss += lossFn(out,label.squeeze())
        outputs.append(out)
        outputs_raw.append(out_raw)
    #print("Validation loss: {}".format(loss)) 
    preds = torch.cat(outputs).detach().cpu().numpy()
    preds_raw = torch.cat(outputs_raw).detach().cpu().numpy()
    labels = torch.cat(labels).detach().cpu().numpy()
    fig, ax = plt.subplots(nrows =3, figsize = (19,10))
    ax[0].plot(preds,label = "Prediction")
    ax[0].plot(labels,label = "Ground Truth")
    plt.legend()
    ax[1].plot(raw_data[:,3],label = "Ticker")
    ax[2].plot(labels,label = "Ground Truth")
    ax[2].plot(preds_raw,label = "Prediction")
    plt.legend()
    plt.savefig("{}/Regression Results {} (MSE: {}).png".format(path,index,loss.item()))
    plt.close()
    

  
  def score_naswot(self,model,loader):
    #nw = naswot(model,loader,self.batch_size,self.cuda_device)
    #return nw.score()
    return 0
  
  def unsup_loss(self, model,loader,binary = False, n_augment = 1):
    loss = 0 
    S = torch.nn.Sigmoid()
    if binary == True:
      lossFn = F.binary_cross_entropy_with_logits
    else:
      lossFn = F.cross_entropy
    samples = len(loader)
    with torch.no_grad(): #disable back prop to test the model
      for i, (x,_) in enumerate(loader):
        x = x.cuda(non_blocking=True, device = self.cuda_device).float()
        for a in range(n_augment):
          x_1 = augment(x,device = self.cuda_device)
          x_2 = augment(x,device = self.cuda_device)
          logits_1 = model(x_1)
          logits_2 = model(x_2)
          loss += lossFn(S(logits_1), S(logits_2))
      averaged_loss = loss/(samples*n_augment)
    return averaged_loss

  def map_to_origin_class( self , labels ):
    c_matrix = np.zeros(shape = (21,2))
    for pred, label,b_labels in zip(self.prediction, labels,self.labels):
      p, l = int(pred[0]),int(label)
      c_matrix[l,p] += 1

    with np.printoptions(linewidth = (10*self.n_classes+20),precision=4, suppress=True):
      pass
      #print(c_matrix)
    return c_matrix 

  def loss_over_sample_size(self,model,dataset):
      n = 2
      u_loss = {}
      u10_loss = {}
      u100_loss = {}
      s_loss = {}
      while n < 16:
        s, u, u10,u100 = self.loss_distribution(model,dataset , sample_per_class = n)
        u_loss[n] = copy.deepcopy(u)
        u10_loss[n] = copy.deepcopy(u10)
        u100_loss[n] = copy.deepcopy(u100)
        s_loss[n] = copy.deepcopy(s)
        n = n*2
      df_supervised = pd.DataFrame(s_loss.items(),columns = ["num samples", "supervised scores"])
      df_supervised = df_supervised.explode("supervised scores")
      df_unsupervised = pd.DataFrame(u_loss.items(),columns = ["num samples", "unsupervised scores"])
      df_unsupervised = df_unsupervised.explode("unsupervised scores")
      df_unsupervised10 = pd.DataFrame(u10_loss.items(),columns = ["num samples", "unsupervised scores"])
      df_unsupervised = df_unsupervised.explode("unsupervised scores")
      df_unsupervised100 = pd.DataFrame(u100_loss.items(),columns = ["num samples", "unsupervised scores"])
      df_unsupervised = df_unsupervised.explode("unsupervised scores")
      df_loss = pd.DataFrame()
      df_loss["samples"] = df_supervised["num samples"]
      df_loss["supervised"] = df_supervised["supervised scores"]
      df_loss["unsupervised"] = df_unsupervised["unsupervised scores"]
      df_loss["unsupervised10"] = df_unsupervised10["unsupervised scores"]
      df_loss["unsupervised100"] = df_unsupervised100["unsupervised scores"]
      return df_loss
  def loss_distribution(self, model, dataset, loss = "all", sample_per_class = 50):
      labels = dataset.get_labels()
      MAX_SIZE = 500
     #print(labels)
      
      strat_sampler = StratifiedBatchSampler(y = labels , batch_size = sample_per_class*21)

      testloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd,batch_sampler = strat_sampler)
      sup_losses = []
      sup10_losses = []
      sup100_losses = []
      for x,y in testloader:
        chunks = -(len(y)//-MAX_SIZE)
        x_i = x.chunk(chunks)
        y_i = y.chunk(chunks)
        subsample = [(x_j,y_j) for x_j, y_j in zip(x_i,y_i)]
           
        s_loss = self.sup_loss(model,subsample)
        s_loss_10 = self.sup_loss(model,subsample,n_augmment = 10)
        s_loss_20 = self.sup_loss(model,subsample,n_augmment = 20)
       #print("loss: {}".format(s_loss))
       #print("loss: {}".format(s_loss_10))
       #print("loss: {}".format(s_loss_20))
        """
        u_loss = self.unsup_loss(model,subsample)
        u10_loss = self.unsup_loss(model,subsample, n_augment = 10)
        u100_loss = self.unsup_loss(model,subsample, n_augment = 100)
        sup_losses.append(s_loss.item())
        """
        sup_losses.append(s_loss.item())
        sup10_losses.append(s_loss_10.item())
        sup100_losses.append(s_loss_20.item())
        
      return sup_losses, sup10_losses, sup100_losses

  def sup_loss(self, model,loader,n_augment = 1 , binary = False):
    loss = 0 
    S = torch.nn.Sigmoid()
    samples = len(loader)
    if binary == True:
      lossFn = F.binary_cross_entropy_with_logits
    else:
      lossFn = F.cross_entropy
    with torch.no_grad(): #disable back prop to test the model
      for i, (x , y) in enumerate(loader):
        for i in range(n_augment):
          x = x.cuda(non_blocking=True, device = self.cuda_device).float()
          y = y.cuda(non_blocking=True, device = self.cuda_device).long()
          x = augment(x,device = self.cuda_device)
          logits = model(x)
          loss += lossFn(S(logits), y)
      averaged_loss = loss/samples
    return averaged_loss

  def c_loss(self,model,loader,n_augment = 1,binary = False):
    loss = 0 
    s_rates = [0.7, 0.7, 0.9,0.7]
    s_hps = [0.15, 0.12, 4,0.49]
    w_rates = [0.3, 0.3, 0.5,0.3]
    w_hps = [0.015, 0.012, 2,0.1]
    w_settings = {"rates": w_rates,"hps":w_hps}
    s_settings = {"rates": s_rates,"hps":s_hps}
    S = torch.nn.Sigmoid()
    if binary == True:
      lossFn = F.binary_cross_entropy_with_logits
    else:
      lossFn = F.cross_entropy
    samples = len(loader)
    with torch.no_grad(): #disable back prop to test the model
      for i, (x,_) in enumerate(loader):
        x = x.cuda(non_blocking=True, device = self.cuda_device).float()
        for a in range(n_augment):
          x_s = augment(x,s_settings,device = self.cuda_device)
          x_w = augment(x,w_settings,device = self.cuda_device)
          logits_w = model(x_w)
          logits_s = model(x_s)
          loss += consistency_loss(logits_s, logits_w)
      averaged_loss = loss/(samples*n_augment)
    return averaged_loss
  def ROC(self,fold = 0):
    fpr , tpr, thresholds = roc_curve(self.labels, self.model_prob) 
    roc_auc = auc(fpr,tpr)
    plt.figure()
    lw = 2
    plt.plot(
        fpr,
        tpr,
        color="darkorange",
        lw=lw,
        label="ROC curve (area = %0.2f)" % roc_auc,
    )
    plt.plot([0, 1], [0, 1], color="navy", lw=lw, linestyle="--")
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("Receiver operating characteristic example")
    plt.legend(loc="lower right")
    plt.savefig("ROC_{}".format(fold))
    return fpr,tpr,thresholds, roc_auc

  def forward_pass(self, model , testloader = None, binary = False, subset = None,n_iter = 1):
    if testloader != None:
      self.testloader = testloader
    elif self.testloader == None and testloader == None:
      raise Exception("No dataloader passed at object initialisation or at forward pass")

    if binary == True:
      s = torch.nn.Sigmoid() #torch.nn.Identity()# torch.nn.Sigmoid()
      self.model_prob = np.zeros(shape = (len(self.testloader)*self.batch_size*n_iter, 1)) # [sample , classes]
    else:
      s = torch.nn.Identity()
      self.model_prob = np.zeros(shape = (len(self.testloader)*self.batch_size*n_iter, self.n_classes)) # [sample , classes]
    self.labels = np.zeros(shape = (len(self.testloader)*self.batch_size*n_iter,1))
    #Pass validation set through model getting probabilities and labels
    with torch.no_grad(): #disable back prop to test the model
      for n in range(n_iter):
        for i, (inputs, labels) in enumerate( self.testloader ):
            start_index = i * self.batch_size + (n * len(self.testloader) * self.batch_size)
            end_index = (i * self.batch_size + (n * len(self.testloader) * self.batch_size)) + self.batch_size
            if self.cuda_device != None:
                inputs = inputs.cuda(non_blocking=True, device = self.cuda_device).float()
            else:
                inputs = inputs.cpu()
            self.labels[start_index:end_index , :] = labels.view(self.batch_size,1).cpu().numpy()
            out = s(model(inputs)).cpu().numpy()
            self.model_prob[start_index:end_index,:] = out
            if subset != None:
              if i > subset:
                self.labels = self.labels[:end_index,:]
                self.model_prob = self.model_prob[:end_index, :]
                break
  def set_cm(self, prediction, labels):
    self.confusion_matrix = confusion_matrix(labels,prediction,labels = list(range(self.n_classes)))
  def update_CM(self):
    self.confusion_matrix += confusion_matrix(self.labels, self.prediction,labels = list(range(self.n_classes))) 

  def balanced_acc(self):
    return balanced_accuracy_score(self.labels, self.prediction) 

  def reset_cm(self):
    self.confusion_matrix = np.zeros(shape = (self.n_classes,self.n_classes)) #Matrix of prediction vs true values

  def predictions_threshold_matrix(self, model_is_binary):
    for i in range(1,20):
      threshold = i * 0.05
      self.predictions(model_is_binary = model_is_binary, THRESHOLD = threshold)
      acc  =  self.T_ACC()
      recall = self.TPR(1)
     #print("Threshold: {} -- Accuracy: {} -- Recall: {}".format(threshold,acc,recall)) 
      self.reset_cm()

  def predictions(self, model_is_binary = False, THRESHOLD = None, no_print = True):
      if model_is_binary:

        self.prediction = np.where(self.model_prob > THRESHOLD, 1,0)
        #for m,p, l in zip(self.model_prob, self.prediction, self.labels):
        # #print("Logit: {} -- Predicted: {} label: {}".format(m,p,l))
        assert self.prediction.shape == (len(self.model_prob),1), "Shape of prediction is {} when it should be {}".format(self.prediction.shape, (len(self.model_prob),1))
      else:
        self.prediction = np.argmax(self.model_prob, axis = 1).reshape(-1,1)
        assert self.prediction.shape == (len(self.model_prob),1),  "Shape of prediction is {} when it should be {}".format(self.prediction.shape, (len(self.model_prob),1))
      self.correct = np.where(self.labels == self.prediction,1,0)

      self.update_CM()
      if not no_print:
          with np.printoptions(linewidth = (10*self.n_classes+20),precision=4, suppress=True):
            #print(self.confusion_matrix)
            pass

  def calculate_loss(self,criterion,model_is_binary = False):
    if model_is_binary:
      return criterion(torch.Tensor(self.model_prob).squeeze(), torch.Tensor(self.labels).squeeze().float())
    else:
      return criterion(torch.Tensor(self.model_prob), torch.Tensor(self.labels).squeeze().long())

  def TP(self, value):
    TP = self.confusion_matrix[value,value]
    return TP
  
  def Correct(self):
    return np.sum(np.diag(self.confusion_matrix))

  def TN(self, value):
    TN = self.Correct() - TP(value)
    return TN 

  def FN(self, value):
    idx = [x for x in range(self.confusion_matrix.shape[0]) if x != value]
    FN = np.sum(self.confusion_matrix[value,idx])
    return FN

  def FP(self, value):
    idx = [x for x in range(self.confusion_matrix.shape[0]) if x != value]
    FP = np.sum(self.confusion_matrix[idx,value])
    return FP

  def P(self,value):
    P = np.sum(self.confusion_matrix[value,:])
    return P

  def N(self,value):
    N = 0
    idx = [x for x in range(self.confusion_matrix.shape[0]) if x != value]
    for i in idx:
      N += np.sum(self.confusion_matrix[i,:])
    return N
  def T(self):
    return np.sum(self.confusion_matrix)
  
  def ACC(self, value):
    return ( self.TP(value) + self.TN(value) ) / ( self.P(value) + self.N(value) )

  def TPR(self, value):
    if self.P(value) >= 1:
      return self.TP(value)/self.P(value)
    else:
      return 0

  def TNR(self, value):
    return self.TN(value)/self.N(value)

  def PPV(self, value):
    tp = self.TP(value)
    return tp/(tp + self.FP(value))

  def NPV(self, value):
    tn = self.TN(value)
    return tn/(tn + self.FN(value))

  def FNR(self, value):
    pass

  def FPR(self, value):
    pass 

  def T_ACC(self) -> float:
    return self.Correct() / self.T()
     












from HPO.searchspaces.utils import graph_joiner, op_joiner
import networkx as nx



def build_bottleneck(stride = 1, step = False):
  graph = [(0,1),(1,2),(2,3),(0,3),(3,4)]
  bottleneck_ops = {
    "0_1_OP": "skip_connect",
    "1_2_OP": "depth_conv_3_1",
    "2_3_OP": "skip_connect",
    "0_3_OP": "skip_connect",
    "3_4_OP": "skip_connect",
  }
  node_ops = {
    "1_activation" : "gelu","1_normalisation": "batch_norm","1_stride": 1,"1_channel_ratio":0.25 if step == 0 else step,
    "2_activation" : "gelu","2_normalisation": "batch_norm","2_stride": stride,"2_channel_ratio":1,
    "3_activation" : "none","3_normalisation": "batch_norm","3_stride": 1,"3_channel_ratio":4,
    "4_activation" : "gelu","4_normalisation": "none","4_stride": 1,"4_channel_ratio":1,
    
  }
  ops = {**bottleneck_ops , **node_ops}
  return graph, ops


def make_layer(blocks, step,stride = 1 ):
  g,op = build_bottleneck(stride = stride,step = step)
  
  for i in range(blocks-1):
    g_i , op_i = build_bottleneck()
    g = graph_joiner(g,g_i)
    op = op_joiner(op,op_i)
  return g , op    
  

def build_resnet( layer_list):
    ops = {"S_0_OP": "max_pool_3_1","0_activation" : "gelu","0_normalisation": "batch_norm","0_stride": 2,"0_channel_ratio":1,"S_activation": None, "S_normalisation": None, "S_channel_ratio": 1, "S_stride": 1, "S_combine" : None}
    graph = [("S",0)]
    
  
    graph_new, ops_new = make_layer(layer_list[0],step = 1 )
    graph = graph_joiner(graph,graph_new)
    ops = op_joiner(ops,ops_new)

    for layer in layer_list[1:]: 
        graph_new, ops_new = make_layer(layer, stride=2,step = 0.5)
        graph = graph_joiner(graph,graph_new)
        ops = op_joiner(ops,ops_new)
    """
        graph_new, ops_new = make_layer(layer_list[2], stride=2,step = 0.5)
        graph = graph_joiner(graph,graph_new)
        ops = op_joiner(ops,ops_new)

    graph_new, ops_new = make_layer(layer_list[3], stride=2,step = 0.5)
    graph = graph_joiner(graph,graph_new)
    ops = op_joiner(ops,ops_new)
    """
    g = nx.DiGraph()
    g.add_edges_from(graph)
    _max = max([ int(i) for i in g.nodes() if str(i).isdigit()])
    ops_out = {}
    for i in ops:
      ops_out[i.replace(str(_max),"T")] =  ops[i]
    for c,i in enumerate(graph):
      if int(i[1]) == int(_max):
        graph[c] = (i[0], "T")
     
    g = nx.DiGraph()
    g.add_edges_from(graph)
    for i in g.nodes():
      ops_out["{}_combine".format(i)] = "ADD"
    return {"graph":graph, "ops":ops_out}

def ResNet50():
    return build_resnet( [3,4,6,3])      

def ResNetDeep():
    return build_resnet( [2,2,1,1,1,1])      

def ResNet101():
    return build_resnet( [3,4,24,3])      



import torch.nn as nn

class MLP(nn.Module):
  def __init__(self, input_size, layers , output_size):
    super().__init__()
    self.stem = nn.Linear(input_size , layers[0])
    self.hidden = nn.ModuleList()
    self.act = nn.ReLU()
    self.outact = nn.Softmax()
    if len(layers) > 1:
      for c,i in enumerate(layers[1:]):
        self.hidden.append(nn.Linear(layers[c], i))
    self.out = nn.Linear(layers[-1], output_size)

  def forward( self, x ):
    x_s = self.act(self.stem(x))
    if len(self.hidden) > 0:
      for i in self.hidden:
        x_s = self.act(i(x_s))
    x_s = self.act(self.out(x_s))
    return x_s
     



import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
def round_average(l : list, t_samples):
  while sum(l) < t_samples:
    l[random.randint(0,len(l)-1)] +=1
  return l
  
class TEPS(Dataset):
  def __init__(self, window_size, train , augmentations, samples_per_class = None,binary = False,one_hot = True,samples_per_epoch = 1,PATH = None,device = None,sub_set_classes = None):
    self.features = 52
    self.device = device
    self.one_hot = one_hot
    self.augmentations = augmentations
    self.sub_set_classes = sub_set_classes
    if binary == True:
      self.n_classes = 2
    else:
      self.n_classes = 21
    if samples_per_class != None and binary == True:
      if self.sub_set_classes != None:
        self.samples_per_class = [0]*20
        for i in self.sub_set_classes:
          self.samples_per_class[i] = int(samples_per_class)
        self.samples_per_class = round_average(self.samples_per_class, samples_per_class)
        #self.samples_per_class = [samples_per_class] + self.samples_per_class
      else:
        self.samples_per_class = [int(samples_per_class/20)]*20
        self.samples_per_class = round_average(self.samples_per_class, samples_per_class)
        self.samples_per_class = [samples_per_class] + self.samples_per_class
    elif samples_per_class != None:
      self.samples_per_class = [samples_per_class]* self.n_classes
    else:
      self.samples_per_class = [0]* self.n_classes
    if PATH == None:
      path = "{}/scripts/datasets/TEPS/split/".format(os.environ["HOME"])
    else:
      path = PATH
    files = os.listdir(path)
    #Get either Training or testing samples
    if train == True:
      filtr = "training"
    else:
      filtr = "testing"
    files_all = [ name for name in files if filtr in name]    
    
    #randomly Subsample to $samples_per_class number of files of each class
    if samples_per_class == None:
      files = files_all
    else:
      files = []
      random.shuffle(files_all)
      for i in files_all:
        if self.sub_set_classes == None:
          if i[6:8].isnumeric():
            if self.samples_per_class[int(i[6:8])] > 0:
              files.append(i)
              self.samples_per_class[int(i[6:8])] -=1
          else:
            if self.samples_per_class[int(i[6])] > 0:
              files.append(i)
              self.samples_per_class[int(i[6])] -=1
          if sum(self.samples_per_class) == 0:
            break
        else:
          if i[6:8].isnumeric():
            if int(i[6:8]) in self.sub_set_classes:
              if self.samples_per_class[int(i[6:8])] > 0:
                print(i)
                files.append(i)
                self.samples_per_class[int(i[6:8])] -=1
          else:
            if int(i[6]) in self.sub_set_classes:
              if self.samples_per_class[int(i[6])] > 0:
                files.append(i)
                self.samples_per_class[int(i[6])] -=1
            if sum(self.samples_per_class) == 0:
              break
    
    data = []
    self.x = {}
    self.y = {}
    self.window = window_size
    for i in files:
      data.append(np.load(path+i))
    self.current_index = 0
    use_all = True
    self.labels = [] 
    self.true_labels = [] 
    for batch in data:
      batch_x = batch[:,1:]
      self.true_labels.append(batch[0,0])
      if binary == True and batch[0,0] > 0:
        batch_y = [1]
      else:
        batch_y = batch[:,0]
      self.add_to_dataset(batch_x,batch_y)
    self.n_features = batch_x.shape[1]
    self.n_samples = self.current_index + 1 #* len(self.augmentations)
    self.samples_per_epoch = samples_per_epoch
    if samples_per_epoch > 1:
      self.true_n_samples = self.n_samples
      self.n_samples *= samples_per_epoch
    self.labels = torch.Tensor(self.labels)
    self.x = torch.stack(list(self.x.values()), dim=0)
    self.y = torch.stack(list(self.y.values()), dim=0)
    print(self.y.shape)
  def add_to_dataset(self,x,y):
    if self.one_hot == True:
      self.x[self.current_index] = torch.from_numpy(np.swapaxes(x,0,1)).float().cuda(device = self.device )
      self.y[self.current_index] = F.one_hot(torch.from_numpy(np.unique(y)).long(),num_classes = self.n_classes).cuda(device = self.device).long()
    else:
      self.x[self.current_index] = torch.from_numpy(np.swapaxes(x,0,1)).float().cuda(device = self.device )
      self.y[self.current_index] = torch.from_numpy(np.unique(y)).cuda(device = self.device).long()
    self.samples_per_class[int(y[0])] += x.shape[0]
    self.current_index += 1
    self.labels.append(y[0])

  def get_n_samples_per_class(self):
    for i in range(self.n_classes):
      print("Samples in class {}: {}".format(i, self.samples_per_class[i]))    

  def set_window_size(self, window_size):
    self.window = window_size
  def get_labels(self):
    return self.labels
  def get_true_labels(self):
    return self.true_labels
  def disable_augmentation(self):
    self.augmentations = False

  def enable_augmentation(self, augmentations):
    self.augmentations = augmentations
  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index
    if self.samples_per_epoch > 1:
      if index >= self.true_n_samples-2:
        index = index % (self.true_n_samples-2)
    x = self.x[index]
    y = self.y[index]
    if self.augmentations:
      for func in self.augmentations:
        x,y = func(x,y)
    if self.one_hot == False:
      return x , y
    return x , y

  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_features

  def __len__(self):
    return self.n_samples -1  


class Train_TEPS(TEPS):

  def __init__(self, window_size = 500, augmentations = False,samples_per_class = None,binary = False,one_hot = False,samples_per_epoch = 1,PATH= None,device = None,sub_set_classes = None): 
    super().__init__(window_size, True, augmentations,samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,PATH = PATH,device = device,sub_set_classes = sub_set_classes)

class Test_TEPS(TEPS):

  def __init__(self, window_size = 500, augmentations = False,samples_per_class = None,binary = False,one_hot = False,samples_per_epoch = 1,device = None,sub_set_classes = None): 
    super().__init__(window_size,  False , augmentations = augmentations,samples_per_class = samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,device = device,sub_set_classes = sub_set_classes)




import HPO.utils.ops1d as ops
import random
#hyperparameter processing
from operator import itemgetter 
import torch.nn as nn
import time
class DataShapeLogger:
  def __init__(self, filename):
    self.filename = filename
    self.log_list = [] 
    MyFile=open(self.filename+".txt",'a') 
    MyFile.write(str(time.time()))
    MyFile.write('\n')
  def log(self, *string):
    string = [str(x) for x in string ]
    out_string = "".join(string)
    MyFile=open(self.filename+".txt",'a')
    
    MyFile.write(out_string)
    MyFile.write('\n')
    MyFile.close()

class DynamicLayer(nn.Module):
  ##Allows for training data batches from datasets with differing numbers of features to be used in the same epoch
  def __init__(self, channels, layer_type):
    super(DynamicLayer,self).__init__()
    self.channels = channels
    self.stem_list = nn.ModuleList()
    self.layer_type = layer_type
    self.stem_index = []
  def forward(self, x):
    features = x.shape[1]
    if features not in self.stem_index:
      self.stem_list.append(self.layer_type(features , self.channels ).cuda())
      self.stem_index.append(features)
    self.last_stem = features
    return self.stem_list[self.stem_index.index(features)](x)


class DynamicStem(DynamicLayer):
  def __init__(self , channels ):
    super().__init__( channels , ops.StdConv)

class DynamicFC(nn.Module):
  def __init__(self , channels, output_dict  , dynamicStem):
    super(DynamicFC,self).__init__()
    self.channels = channels
    self.stem_list = nn.ModuleList()
    self.layer_type = nn.Linear
    self.stem_index = []
    self.output_dict = output_dict
    self.dynamicStem = dynamicStem
  def forward(self, x):
    features = self.dynamicStem.last_stem
    if features not in self.stem_index:
      self.stem_list.append( self.layer_type( self.channels , self.output_dict[features] ).cuda() )
      self.stem_index.append( features )
    return self.stem_list[self.stem_index.index(features)](x)


class Model(nn.Module):
  def __init__(self, input_size, output_size, hyperparameters, one_fc_layer = False , dynamicStem = False , dynamicFC = False , output_dict = None):
    super(Model,self).__init__()
    self.one_fc_layer = one_fc_layer
    self.log_flag = True
    self.p = 0
    self.logger = DataShapeLogger("logger.txt")
    self.hyperparameters = hyperparameters  
    self.channels = hyperparameters["channels"]
    self.outputs = output_size
    if output_dict != None:
      self.output_dict = output_dict
    self.normal_cells = nn.ModuleList()
    self.reduction_cells = nn.ModuleList()
    self.layers = hyperparameters["layers"]
    if dynamicStem == True:
      self.in_conv = DynamicStem(self.channels)
    else:
      self.in_conv = ops.StdConv(input_size[0], self.channels)
    self.build_cells(hyperparameters)
    self.gap = ops.AdaAvgPool() 
    self.fc_list = nn.ModuleList()
    channels = self.channels
    if self.one_fc_layer == True:
      while channels > 2*output_size:
        self.fc_list.append(nn.Linear(channels,channels//2))
        self.fc_list.append(nn.ReLU())
        channels = channels // 2

    if output_dict != None and dynamicFC != False:
      self.output_dict = output_dict
      self.fc = DynamicFC(self.channels , output_dict, self.in_conv)

    else:
      self.fc = nn.Linear(channels, output_size)
    self.outact = nn.Softmax(dim = 1)
    self.outact_eval = nn.Softmax(dim = 0)
    

  def reset_stem(self,in_features : int):
    self.in_conv = ops.StdConv(in_features, self.channels)
  def get_channels(self):
    return self.channels
  def reset_fc(self, output_size : int ):
    self.fc_list = nn.ModuleList()
    channels = self.channels
    if self.one_fc_layer == True:
      while channels > 2*output_size:
        self.fc_list.append(nn.Linear(channels,channels//2))
        self.fc_list.append(nn.ReLU())
        channels = channels // 2
    self.fc = nn.Linear(channels, output_size)

  def _build_dict(self,parameters : dict, keyword : str):
    _dictionary = dict()
    keyword_length = len(keyword)
    id_index = keyword_length + 1
    
    for parameter in parameters:
      if parameter[:keyword_length] == keyword:
        cell_id = int(parameter[id_index])

        operation_key = parameter[id_index + 2 : ]
        operation_value = parameters[ parameter ]
        
        if cell_id in _dictionary.keys():        
          _dictionary[ cell_id ][ operation_key ] = operation_value
         
        else: #if dictionary doesnt exist, make it
          _dictionary[ cell_id ] = { operation_key : operation_value }

    return _dictionary
  
  def build_cells(self, parameters): 
    conv_dictionary = self._build_dict(parameters, "normal_cell")
    redu_dictionary = self._build_dict(parameters, "reduction_cell")
    
    for i in range(parameters["layers"]):
         
      self.normal_cells.append(Cell(conv_dictionary[1],self.channels,self.channels,p = self.p))
      if i < (self.layers -1):
        self.reduction_cells.append(Cell(redu_dictionary[1],self.channels,channels_out = self.channels*1, p = self.p))
        #self.channels*=2 
  
  def _forward(self,x):
    x = self.in_conv(x)
    for i in range(self.layers):
      x = self.normal_cells[i](x) 
      if i != (self.layers -1):
        x = self.reduction_cells[i](x)
    x = self.gap(x)
    x = x.squeeze()
    if self.one_fc_layer == True:
      for i in self.fc_list:
        x = i(x)
    if len(x.shape) == 1:
      x = x.view(1, -1)
    #x = self.outact(x)
    return x  

  def _forward_eval(self,x):
    x = self.in_conv(x)
    for i in range(self.layers):
      x = self.normal_cells[i](x) 
      if i != (self.layers -1):
        x = self.reduction_cells[i](x)
    x = self.gap(x)
    x = x.squeeze()
    if self.one_fc_layer == True:
      for i in self.fc_list:
        x = i(x)
    x = self.fc(x)

    x = self.outact_eval(x)
    return x  

  def forward(self,x):
    x=self._forward(x)
    return self.outact(self.fc(x))

class Ops(nn.Module):
  def __init__(self, parameters, channels_in,channels_out, p):
    super(Ops,self).__init__()
    self.args = {}
    self.channels_in = channels_in
    self.channels_out = channels_out
    self.multicompute = False
    self.p = p
    self.input = []
    for i in parameters:
      if i == "type":
        self.op = parameters[i]
        self.operation = self.get_operation(parameters[i])
      elif i[:-2] == "input":
          if parameters[i] not in self.input:
            self.input.append(parameters[i])
      else:
          self.args[i] = parameters[i]
    self.compute = nn.ModuleList()
    if len(self.input) > 1:
      self.compute.append(ops.StdAdd())
      self.multicompute = True
      self.pool = nn.AvgPool1d(2)
    self.compute.append(self.operation(**self.args))
  def get_required(self) -> list:
    return self.input

  def get_operation(self, op_key):
    if op_key == "StdConv":
      operation = ops.StdConv
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
      self.args["padding"] = "same" 
    elif op_key == "SepConv3":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 3
      self.args["stride"] = 1
      self.args["padding"] = "same"
      self.args["separable"] = True
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "SepConv5":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 5
      self.args["stride"] = 1
      self.args["padding"] = "same" 
      self.args["separable"] = True
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "SepConv7":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 7
      self.args["stride"] = 1
      self.args["padding"] = "same" 
      self.args["separable"] = True
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "Conv3":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 3
      self.args["stride"] = 1
      self.args["padding"] = "same"
      self.args["separable"] = False
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "Conv5":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 5
      self.args["stride"] = 1
      self.args["padding"] = "same" 
      self.args["separable"] = False
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "Conv7":
      operation = ops.ConvBranch
      self.args["kernel_size"] = 7
      self.args["stride"] = 1
      self.args["padding"] = "same" 
      self.args["separable"] = False
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "MaxPool5":    
      operation = ops.Pool
      self.args["pool_type"] = "max"
      self.args["kernel_size"] = 5
      self.args["padding"] = 2
    elif op_key == "MaxPool7":    
      operation = ops.Pool
      self.args["pool_type"] = "max"
      self.args["kernel_size"] = 7
      self.args["padding"] = 3
    elif op_key == "AvgPool5":    
      operation = ops.Pool
      self.args["pool_type"] = "avg"
      self.args["kernel_size"] = 5
      self.args["padding"] = 2
    elif op_key == "AvgPool7":    
      operation = ops.Pool
      self.args["pool_type"] = "avg"
      self.args["kernel_size"] = 7
      self.args["padding"] = 3
    elif op_key == "Identity":    
      operation = ops.Identity
    elif op_key == "FactorizedReduce":
      operation = ops.FactorizedReduce
      self.args["C_in"] = self.channels_in 
      self.args["C_out"] = self.channels_out
    elif op_key == "":
      operation = ops.StdConv
    return operation
  def forward( self, x ):
    #print(self.op)
    for count,i in enumerate(self.compute):
      if self.multicompute and count == 0:
        #print("Size Before operation: ", x[0].size(),x[1].size())
        x = i(*x) 
        #print("Size After operation: ", x.size())
      else:
        #x = self.dropout(x)
        x = i(x)
        #print("Size After operation: ", x.size())
    #if self.multicompute:
      #x = self.pool(x)
    return x 
  
  def process( self, x : dict):
    return self.forward(itemgetter( *self.get_required() )( x ))  

class Cell(nn.Module):
  """
  Contains a series of operations and information links
  """
  def __init__(self,parameters,channels_in,channels_out,p):
    super(Cell, self).__init__()
    self.ops_id = [] #numerical identifier for each operation 
    self.ops = []
    self.p = p
    self.channels_in = channels_in
    self.channels_out = channels_out
    self.inputs = []

    self.build_ops(parameters)
    self.output_operation = max(self.ops_id) #parameters["num_ops"]
    self.compute_order = nn.ModuleList()
    self.cal_compute_order()
  def _build_dict(self,parameters : dict, keyword : str):
    _dictionary = dict()
    keyword_length = len(keyword)
    id_index = keyword_length + 2
    
    for parameter in parameters:
      if parameter[:keyword_length] == keyword:
        cell_id = int(parameter[id_index])

        operation_key = parameter[id_index + 2 : ]
        operation_value = parameters[ parameter ]
        
        if cell_id in _dictionary.keys():        
          _dictionary[ cell_id ][ operation_key ] = operation_value
         
        else: #if dictionary doesnt exist, make it
          _dictionary[ cell_id ] = { operation_key : operation_value }

    return _dictionary
  def build_ops(self, parameters):
    ops_dictionary = self._build_dict(parameters, "op")  
    for i in ops_dictionary:
      
      self.ops.append(Ops(ops_dictionary[i], self.channels_in,self.channels_out, self.p))
      self.ops_id.append(i)
  def cal_compute_order(self):
    #Calculate the order in which the operations in a cell should be computed
    self.com_order_id_list = [] #Tracks the ID of the operations

    current_operation = self.output_operation 
    self._compute_order(current_operation)
    
    ml = nn.ModuleList()
    self.new_compute_order = []
    for i in range(len(self.compute_order)-1,-1, -1):

      if self.compute_order[i] not in ml:
       
        self.new_compute_order.append(self.com_order_id_list[i]) # The final operation order ID list
        ml.append(self.compute_order[i])
    self.compute_order = ml



  def _compute_order(self,operation): 
      #Appends an operation to the compute list and then traverses the prerequisite 
      #operations recursively 
      self.com_order_id_list.append(operation)
      self.compute_order.append(self.ops[self.ops_id.index(operation)])


      #Get the requirements compute inputs as long as the input is not zero
      for i in self.ops[self.ops_id.index(operation)].get_required():
        if i != 0:
          self._compute_order(i)
      return 
    
    
  def forward(self, x):
    outputs = {0 : x}
    for id,op in zip(self.new_compute_order,self.compute_order):
      outputs[id] = op.process(outputs)

    return outputs[self.output_operation] 

 



import numpy as np 
import torch 
import random 
import json


def set_seed(JSON_PATH):
    with open(JSON_PATH, "r") as f:
        data = json.load(f)

    seed = random.randint(0,999)
    data["SEED"] = seed
    with open(JSON_PATH, "w") as f:
        json.dump(data, f)
    random.seed(seed)
    torch.manual_seed(seed)
    np.random.seed(seed)



from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace
from HPO.algorithms.algorithm_utils import train_eval
import json
def main(worker, configspace : ConfigurationSpace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]
    
  train = train_eval( worker , json_config)
  configs = configspace.sample_configuration(SETTINGS["TOTAL_EVALUATIONS"] - train.ID_INIT)
  scores ,recall , pop= train.eval(configs)
  #print("Best Score: ", max(scores))      
  #plot_scores(scores)
  
  best_score = max(scores)
  best_config = pop[scores.index(max(scores))]
  best_rec = recall[scores.index(max(scores))]
 



import argparse
import pandas as pd
import os
import subprocess
import psutil

global GPU_UUID_MAP
GPU_UUID_MAP = dict()

def get_current_cpu_mem_usage(field='rss', process=None):
    ''' Return the memory usage in MB of provided process
    '''
    if process is None:
        process = psutil.Process(os.getpid())
    mem = getattr(process.memory_info(), field)
    mem_MiB = mem / float(2 ** 20)
    return mem_MiB

def sanitize_single_line_output_from_list_gpu(line):
    """ Helper to parse output of `nvidia-smi --list-gpus`
    
    Args
    ----
    line : string
        One line from `nvidia-smi --list-gpus`
        
    Returns    
    -------
    ldict : dict
        One field for each of name, num, and uuid
    
    Examples
    --------
    >>> s = "GPU 0: Tesla P100-PCIE-16GB (UUID: GPU-4b3bcbe7-8762-7baf-cd29-c1c51268360d)"
    >>> ldict = sanitize_single_line_output_from_list_gpu(s)
    >>> ldict['num']
    '0'
    >>> ldict['name']
    'Tesla-P100-PCIE-16GB'
    >>> ldict['uuid']
    'GPU-4b3bcbe7-8762-7baf-cd29-c1c51268360d'
    """
    num, name, uuid = map(str.strip, line.split(":"))
    num = num.replace("GPU ", "")
    name = name.replace(" (UUID", "").replace(" ", "-")
    uuid = uuid.replace(")", "")
    return dict(num=num, name=name, uuid=uuid)

def lookup_gpu_num_by_uuid(uuid):
    ''' Helper method to lookup a gpu's integer id by its UUID string
    
    Args
    ----
    uuid : string
    
    Returns
    -------
    int_id : int
        Integer ID (matching index of PCI_BUS_ID sorting used by nvidia-smi)
    '''
    global GPU_UUID_MAP
    try:
        return GPU_UUID_MAP[uuid]
    except KeyError:
        result = subprocess.check_output(['nvidia-smi', '--list-gpus'])
        # Expected format of 'result' is a multi-line string:
        #GPU 0: Tesla P100-PCIE-16GB (UUID: GPU-4b3bcbe7-8762-7baf-cd29-c1c51268360d)
        #GPU 1: Tesla P100-PCIE-16GB (UUID: GPU-f9acf3b8-b5fa-31c5-ecce-81add0ee6a3e)
        #GPU 2: Tesla V100-PCIE-32GB (UUID: GPU-89d98666-7ceb-ccde-136f-28e562834116)

        # Convert into a dictionary mapping a UUID to a plain GPU integer id
        row_list = [
            sanitize_single_line_output_from_list_gpu(line)
            for line in result.decode('utf-8').strip().split('\n')]
        GPU_UUID_MAP = dict(zip([d['uuid'] for d in row_list], [d['num'] for d in row_list]))
        return GPU_UUID_MAP.get(uuid, None)
    
def get_current_cpu_and_gpu_mem_usage_df(query_pids=None):
    """ Get a dataframe of the current cpu and gpu memory usage.

    Will assess usage for either a provided list of processes,
    or the current process and any child process.

    Works only on Unix systems (Linux/MacOS). Not tested on Windows!

    Args
    ----
    query_pids : list or None
        If provided, each entry in list should be an integer process id

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """

    # Determine list of process ids to track called "keep_pids"
    if query_pids is None:
        keep_pids = []
        current_process = psutil.Process(os.getpid())
        keep_pids.append(current_process.pid)
        for child_process in current_process.children(recursive=True):
            keep_pids.append(child_process.pid)
    elif isinstance(query_pids, list):
        keep_pids = list(map(int, query_pids))
    else:
        keep_pids = list(map(int, query_pids.split(',')))

    # Obtain dataframe of memory usage for each process
    cpu_row_list = list()
    for pid in keep_pids:
        proc = psutil.Process(pid)
        mem_val = get_current_cpu_mem_usage(process=proc)
        row_dict = dict()
        row_dict['pid'] = pid
        row_dict['cpu_mem'] = mem_val
        cpu_row_list.append(row_dict)
    cpu_usage_by_pid_df = pd.DataFrame(cpu_row_list)

    # Request dataframe of GPU memory usage from NVIDIA command-line tools
    result = subprocess.check_output(
        [
            'nvidia-smi',
            '--query-compute-apps=pid,gpu_name,gpu_uuid,process_name,used_memory',
            '--format=csv,nounits,noheader'
        ])
    # Convert lines into a dictionary
    keys = ['pid', 'gpu_name', 'gpu_uuid', 'process_name', 'used_memory']
    row_list = [
        dict(zip(keys, map(str.strip, x.split(',')))) for x in result.decode('utf-8').strip().split('\n')]

    all_usage_df = pd.DataFrame(row_list)
    all_usage_df['pid'] = all_usage_df['pid'].astype(int)
    all_usage_df['gpu_mem'] = all_usage_df['used_memory'].astype(float)
    gpu_usage_by_pid_df = all_usage_df[all_usage_df['pid'].isin(keep_pids)].copy()
    gpu_usage_by_pid_df['gpu_id'] = [int(lookup_gpu_num_by_uuid(v)) for v in gpu_usage_by_pid_df['gpu_uuid'].values]
    del gpu_usage_by_pid_df['gpu_uuid']
    del gpu_usage_by_pid_df['used_memory']
    # Combine GPU and CPU into one df
    row_dict = dict()
    row_dict['cpu_mem'] = cpu_usage_by_pid_df['cpu_mem'].sum()
    total = 0.0
    for gpu_id in map(int, GPU_UUID_MAP.values()):
        mem_val = gpu_usage_by_pid_df.query("gpu_id == %d" % gpu_id)['gpu_mem'].sum()
        row_dict['gpu_%d_mem' % gpu_id] = mem_val
        total += mem_val
    row_dict['gpu_total_mem'] = total
    agg_df = pd.DataFrame([row_dict])
    return agg_df, cpu_usage_by_pid_df, gpu_usage_by_pid_df
    

if __name__ == '__main__':
    # Good habit for any code using CUDA
    # Makes sure CUDA_VISIBLE_DEVICES ids align with nvidia-smi
    # since nvidia-smi sorts by pci_bus location of the GPU
    os.environ['CUDA_DEVICE_ORDER'] = "PCI_BUS_ID"

    import numpy as np
    import time
    try:
        import tensorflow as tf
        HAS_TF = True
    except Exception:
        HAS_TF = False

    if HAS_TF:
        gpus = tf.config.experimental.list_physical_devices('GPU')
        if gpus:
            try:
                # Currently, memory growth needs to be the same across GPUs
                for gpu in gpus:
                    tf.config.experimental.set_memory_growth(gpu, True)
                #logical_gpus = tf.config.experimental.list_logical_devices('GPU')
                print("Configured GPUs with memory growth")
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

    pd.set_option('display.precision', 3)
    parser = argparse.ArgumentParser()
    parser.add_argument('--include_np_arrays', default=1, type=int)
    parser.add_argument('--include_tf_cpu_arrays', default=1, type=int)
    parser.add_argument('--include_tf_gpu_arrays', default=1, type=int)
    parser.add_argument('--query_pids', default=None, type=str)
    args = parser.parse_args()

    max_step = 6
    all_arrays = list() # make arrays persist so wont be garbage collected

    for step in range(max_step):
        print("--- Begin step %d" % step)

        if step > 0 and step < max_step//2:
            if args.include_np_arrays:
                A = np.random.randn(1000000, 64).astype(np.float32)
                print("Allocated numpy float64 arr of shape (%d,%d) and size %.2f MB" % (A.shape[0], A.shape[1], A.nbytes / (2**20)))
                all_arrays.append(A)

            if HAS_TF:
                if args.include_tf_cpu_arrays:
                    with tf.device("/device:cpu:0"):
                        tfA = tf.multiply(A, 1.0)
                        print("Allocated tf float64 arr of shape (%d,%d) and size %.2f MB on device %s" % (
                            tfA.shape[0], tfA.shape[1], A.nbytes / (2**20), tfA.device))
                        all_arrays.append(tfA)

                if args.include_tf_gpu_arrays:
                    for localid, gpu_num in enumerate(os.environ.get('CUDA_VISIBLE_DEVICES', '').split(',')):
                        with tf.device('/device:gpu:%s' % localid):
                            tfB = tf.multiply(A, 1.0)
                            tfB += 1 # force gpu to do some work
                            print("Allocated tf float64 arr of shape (%d,%d) and size %.2f MB on device %s" % (
                                tfB.shape[0], tfB.shape[1], A.nbytes / (2**20), tfB.device))
                            all_arrays.append(tfB)

        if step == 1:
            n_per_step = len(all_arrays) # Num arrays allocated each step

        if step > max_step // 2:
            for _ in range(n_per_step):
                arr = all_arrays.pop()
                print("Deleting most recent arr of type %s" % type(arr))
                del arr

        # Wait a bit for garbage collection etc
        time.sleep(0.5)

        print("Reporting current memory")
        usage_df, cpu_pid_df, gpu_pid_df = get_current_cpu_and_gpu_mem_usage_df(query_pids=args.query_pids)
        print(usage_df)

        print("--- End step %d" % step)



import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt


class CIFAR100(Dataset):
  def __init__(self,name,device,augmentation = False):
    self.PATH = "{}{}".format("/home/cmackinnon/scripts/datasets/cifar-100-python/",name)
    self.augmentation = augmentation
    ##LOAD SAMPLES AND LABELS FROM .npy FILE
    def unpickle(file):
        import pickle
        with open(file, 'rb') as fo:
            dict = pickle.load(fo, encoding='bytes')
        return dict
    data_train = unpickle(self.PATH)
    #LOAD IMAGES FROM FILE AND RESHAPE TO [SAMPLES, CHANNELS, ROWS, COLUMNS]
    self.x = torch.from_numpy(
        data_train[b'data'].reshape(-1,3,32,32)
        ).cuda(device = device).float()
    #LOAD LABELS
    self.y = torch.from_numpy(
        np.asarray(data_train[b'fine_labels'] )
        ).cuda(device = device)
    self.n_classes = len(torch.unique(self.y))
    self.n_features = self.x.shape[1]
  def __getitem__(self,index):
    x ,y = self.x[index], self.y[index]
    if self.augmentation:
      for f in self.augmentation:
        x,y = f(x,y)
    return x,y.long()
  def __len__(self):
    return len(self.y)

  def get_n_classes(self):
    return self.n_classes
    
  def get_n_features(self):
    return self.n_features
class CIFAR100_Train(CIFAR100):
  def __init__(self,device,**kwargs):
    name ="train"
    super(CIFAR100_Train,self).__init__(name = name, device = device,**kwargs)
    
class CIFAR100_Test(CIFAR100):
  def __init__(self,device,**kwargs):
    name = "test"
    super(CIFAR100_Test,self).__init__(name = name, device = device,**kwargs)





import torch.nn as nn
import torch

class layer(nn.Module):
  def __init__(self, idx , channels , kernels):
    super(layer, self).__init__()
    self.conv = nn.Conv1d(channels[idx],channels [idx + 1], kernels[idx])
    self.bn = nn.BatchNorm1d(channels[idx+1])
    self.act = nn.ReLU()

  def forward(self , x):
    out = self.conv(x)
    out = self.bn(out)
    return self.act(out)


class FCN(nn.Module):
  def __init__(self, input_size , channels = [64 , 64, 64] , kernels = [8 , 5, 3]  , classes = 2):
    super(FCN,self).__init__()
    self.channels = channels
    self.stem = nn.Conv1d(input_size , channels[0], 1 )
    self.layers = nn.ModuleList()
    self.GAP = nn.AdaptiveAvgPool1d(1)
    self.fc = nn.Linear(channels[-1], classes)
    self.act = nn.ReLU()
    self.fcact = nn.Softmax()
    for idx in range(len(channels[:-1])):
        self.layers.append(layer(idx , channels , kernels))
 
  def get_channels(self):
    return self.channels[-1]

  def _forward(self , x ):
    x = self.stem(x)
    for i in self.layers:
      x = i.forward(x)
    x = self.GAP(x)
    return torch.flatten(x, 1)
  
  def forward(self, x):
    x = self._forward(x)
    return self.fcact(self.fc(x))



import torch
import random
from torch.nn.functional import interpolate
import time
import numpy as np
from operator import itemgetter
from scipy.interpolate import CubicSpline
def time_test(func, n = 100, batch_size = 10 , window_length = 1000, features = 27):
    start = time.time()
    x = torch.rand((batch_size, window_length, features))

    for i in range(n):
        out = func(x)
    print("Total time for ",func.__name__,": ", time.time()- start , " Seconds")




def jitter(x : torch.Tensor, y,sigma=0.03, device = None):
    
    n = torch.distributions.normal.Normal(loc=0., scale=sigma)
    return torch.add(x,n.sample(x.shape).cuda(device = device)),y 


def scaling(x : torch.Tensor, y,sigma=0.05, device = None):

    # https://arxiv.org/pdf/1706.00527.pdf
    n = torch.distributions.normal.Normal(loc=1., scale=sigma)
    s = n.sample((x.shape[0],x.shape[2])).cuda(device = device)
    return torch.mul(x, s[:,None,:]),y


def rotation(x : torch.Tensor,y):
    flip = torch.randint(0,2,size = (x.shape[0],x.shape[2]))
    flip = torch.where(flip != 0, flip, -1)
    rotate_axis = torch.arange(x.shape[2])
    s = np.random.shuffle(np.array(range(x.shape[2])))
    rotate_axis[np.array(range(x.shape[2]))] = rotate_axis[s].clone()
    return flip[:,None,:] * x[:,:,rotate_axis],y



def permutation(x : torch.Tensor,y, max_segments=5, seg_mode="equal"):
    #make array of ints from 1 to window_length
    num_seqments_per_batch = torch.randint(0 , max_segments, 
        size = (x.shape[0],))
    #loop through batches and split into the 
    #defined number of segments for each 
    for idx, sequence in enumerate(x):
        if num_seqments_per_batch[idx] > 1:

            #returns a list of tensors
            chunk_list = torch.chunk(sequence, num_seqments_per_batch[idx], dim = 0)
            #Shuffle
            shuffled_chunks = itemgetter(*np.random.permutation(len(chunk_list)))(chunk_list)
            permutated_sequence = torch.cat(shuffled_chunks, dim = 0)
            x[idx] = permutated_sequence


    return x,y






def magnitude_warp(x : torch.Tensor,y, sigma=0.2, knot=4):
    def h_poly(t):
        tt = t[None, :]**torch.arange(4, device=t.device)[:, None]
        A = torch.tensor([
            [1, 0, -3, 2],
            [0, 1, -2, 1],
            [0, 0, 3, -2],
            [0, 0, -1, 1]
        ], dtype=t.dtype, device=t.device)
        return A @ tt


    def interp(x, y, xs):
        m = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
        m = torch.cat([m[[0]], (m[1:] + m[:-1]) / 2, m[[-1]]])
        idxs = torch.searchsorted(x[1:], xs)
        dx = (x[idxs + 1] - x[idxs])
        hh = h_poly((xs - x[idxs]) / dx)
        return hh[0] * y[idxs] + hh[1] * m[idxs] * dx + hh[2] * y[idxs + 1] + hh[3] * m[idxs + 1] * dx

    orig_steps= torch.linspace(0, x.shape[1]-1, x.shape[1])
    n = torch.distributions.normal.Normal(loc=1., scale=sigma)
    random_warps = n.sample((x.shape[0],knot+2 , x.shape[2]))
    warp_steps = (torch.ones((x.shape[2],1))*(torch.linspace(0, x.shape[1]-1., knot+2))).T
    ret = torch.zeros(x.shape)
    warper = torch.zeros(orig_steps.shape[0],x.shape[2])
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            warper[:,dim] =  interp(warp_steps[:,dim], random_warps[i,:,dim],  orig_steps)

        ret[i] = pat * warper

    return ret,y

def crop(x : torch.Tensor,y, crop_min = 0.85, crop_max = 0.95, device = None):
  sig_len = x.shape[2]
  length= random.uniform(crop_min,crop_max)
  length = int(length * sig_len)
  if random.choice([0,1]) == 1:
    return  x[:,:,:length],y
  else:
    return  x[:,:,(sig_len-length):],y

def window_warp(x : torch.Tensor,y, num_warps = 3, ratios = [0.5,0.75,2],size = 0.3, device = None):
    for i in range(num_warps):
        start = random.randint(1, x.shape[2]-10) 
        end = min([x.shape[2],start+random.randint(2, int(x.shape[2]*size))])
        out= interpolate(x[:,:,start:end], scale_factor = random.choice(ratios)).cuda(device = device)
        x = torch.cat((x[:,:,:start],out,x[:,:,end:]),dim = 2)
    if x.shape[2] > 10000:
        print("Window Warp exceeded size: {}".format(x.shape[2]))
        x = x[:,:,-10000:]

    return x,y

def cutout(x,y, perc=.1, device = None):    
    seq_len = x.shape[2]    
    win_len = int(perc * seq_len)    
    start = np.random.randint(0, seq_len-win_len-1)    
    end = start + win_len    
    start = max(0, start)    
    end = min(end, seq_len)    
    #print("[INFO] start={}, end={}".format(start, end))    
    x[:,:,start:end] = 0    
    # return new_ts, ts[start:end, ...]    
    return x,y 

def mix_up(x,y, m = 0.3, device = None):
    dist = torch.distributions.beta.Beta(m,m)
    DEBUG = False
    #x = [batch, channels , length]
    #y = [batch, 1]
    for i in range(x.shape[0]):
      i_2 = list(range(x.shape[0]))
      i_2.remove(i)
      i_2 = random.choice(i_2)
      mix = dist.sample()
      MTS_1 = x[i,:,:].clone()
      MTS_2 = x[i_2,:,:].clone()
      LAB_1 = y[i]
      LAB_2 = y[i_2]
      x[i,:,:] = (MTS_1 * mix)  + (MTS_2 *(1-mix))
      y[i] = (LAB_1 * mix) + (LAB_2 * (1-mix))
      if DEBUG == True:
        print("mix value is {} and label value is {}".format(mix,y[i]))
        plt.plot(MTS_1[26,:].cpu(),label = "series 1",alpha = 0.7)
        plt.plot(MTS_2[26,:].cpu(),label = "series 2",alpha = 0.7)
        plt.plot(x[i,26,:].cpu(),label = "Mix",alpha = 0.7)
        plt.legend()
        plt.show()
    return x ,y

def cut_mix(x,y, perc=.1, device = None):    
    seq_len = x.shape[2]    
    win_len = int(perc * seq_len)    
    start = np.random.randint(0, seq_len-win_len-1)    
    end = start + win_len    
    for i in range(x.shape[0]):    
        start = max(0, start)    
        end = min(end, seq_len)    
        x[i,:,start:end] = x[random.randint(0,x.shape[0]-1),:,start:end]    
    # return new_ts, ts[start:end, ...]    
    return x,y 
if __name__ == "__main__":

    import torch.nn as nn
    from torch import Tensor
    from HPO.data.teps_datasets import Train_TEPS
    from torch.utils.data import DataLoader
    import random
    import timeit
    import matplotlib.pyplot as plt
    jitter = Jitter(device = 0)
    scaling = Scaling(device = 0)
    funcs = [crop,jitter,jitter_func,scaling_func,scaling,window_warp]
    batch_size = 32
    window_length = 500
    features = 27
    train_dataset = Train_TEPS()
    train_dataloader = DataLoader( train_dataset, batch_size=batch_size,
      shuffle = True,drop_last=True)
    for i,(s,l) in enumerate(train_dataloader):
        x = s.cuda()
        y = l.cuda()
        if i > 0:
          break
        for func in funcs:
           
            """
            print(x.shape)
            print(func)
            #plt.plot(x[0,26,:].cpu(),label = "orig")
            x_p,y_p = func(x,y,device = 0)
            x_p = x_p[0,26,:].cpu()
            #plt.plot(x_p ,alpha = 0.5,label = "aug")
            #plt.legend()
            #plt.show()
            """
            print("from __main__ import {}, {}, {}".format(func.__name__, "x","y"))
            t = timeit.timeit("{}(x,y)".format(func.__name__), "from __main__ import {}, {}, {}".format(func.__name__, "x","y"), number = 1000)
            print("Total time for ",func.__name__,": ",t , " Seconds")
            print("Time per sample: {}".format(t/batch_size))




from multiprocessing import Pool
from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace


def main(worker, configspace : ConfigurationSpace):

  TOTAL_EVALUATIONS = 500
  cores = 1
  pop = []
  scores = [] 


  models = os.listdir("/home/snaags/scripts/HPO/src/HPO/model_zoo/")
  models = sorted(models)
  print(models[-2])
  model = load_model(models[-2])
  dataset = generate_soft_labels( model )

  if cores == 1:
    results = []
    completed_configs = []
    for i in range(TOTAL_EVALUATIONS):
      config = configspace.sample_configuration()
      results.append(worker(config))
      completed_configs.append(config)
      with open("Random.csv", "w") as csvfile:
        writer = csv.writer(csvfile)
        for config,score in zip(completed_configs, results):
          writer.writerow([score, config]) 

  else:
    population = configspace.sample_configuration(TOTAL_EVALUATIONS)
    pop = []
    for i in population:
      pop.append(i.get_dictionary())
    with Pool(processes = cores) as pool:
        results = pool.map(worker, pop)
    for i in results:
      scores.append(i)
    with open("Random.csv", "w") as csvfile:
      writer = csv.writer(csvfile)
      for config,score in zip(pop, scores):
        writer.writerow([config,score]) 
  
  print("Best Score: ", max(scores))      
  plot_scores(scores)
  
  best_config = pop[scores.index(max(scores))]
  best_score_validate = worker.validate(best_config)   
  print(best_config)
  print(best_score_validate) 
 


import csv
import math
import os 
import copy
import torch
from multiprocessing import Pool
import time
import collections
import random
import matplotlib.pyplot as plt 
import numpy as np
from ConfigSpace.configuration_space import Configuration
import ConfigSpace.util as csu
from HPO.utils.visualisation import plot_scores
from HPO.algorithms.algorithm_utils import train_eval 
###Tounament Selection/Aging Selection
  #
  #parameters:
  # S : (int) number of solutions to sample at each iteration
  # P : population size
  # C : Total cycles of evolution 
  # parent : highest-accuracy model in sample
  # history : Record of all models
  # population : Currently active models


# IDEAS
#
# - Run aging evolution for "species" with 1 species per core allowing cross-over occationally
#g
#
#
#
#
#

def F1(acc,rec):
    b = 1
    if acc != None and rec != None:
      acc += 0.0001
      rec += 0.0001
      
    return (1+b**2)+ ((acc*rec)/(((b**2)*(acc**-1))+(rec**-1)))

class Model:
  MUTATE_RATE = 0.2
  def __init__(self, cs):
    self._arch = None
    self.accuracy = None
    self.cs = cs
    self.recall = None

  def set_arch(self, arch):
    self._arch = arch
  def exchange_one( self , key : str ):
    new_hp = self.cs.sample_configuration().get_dictionary()[key]
    
    self._arch[key] = new_hp
  
  def __repr__(self):
    return "Model(Accuracy: {} , Recall: {})".format(self.accuracy,self.recall)
  def perturb(self, key : str):
    value_not_valid = True
    value_type = self.fix_type(key)
    while value_not_valid:
      try:
        self._arch[key] = self._arch[key] + value_type(self._arch[key]*random.uniform(0,Model.MUTATE_RATE)*random.choice([-1,1]))
      except ValueError:
        value_not_valid = True
      else:
        value_not_valid = False

  def fix_type(self,key):
    hp = str(self.cs[key])
    if "Integer" in hp:
      return round
    if "Float" in hp:
      return float
    else:
      return str


  def set_value(self, name, value):
    self._arch[name] = value
  def get_params(self):
    return list(self.arch().keys())
  def arch(self):
    return self._arch

def train_and_eval_population(worker, population, sample_batch, train):
  configs = []
  for model in population:
    if model.accuracy == None:
      configs.append(model.arch())
  acc , _rec , _config = train.eval( configs )
  for mod,result,recall in zip(population, acc, _rec):
    mod.accuracy = result
    mod.recall = recall
def model_change(parent, child):
  p_dict , c_dict = parent.arch().get_dictionary(), child.arch().get_dictionary()
  for _,i in zip(p_dict, c_dict):
    if p_dict[i] != c_dict[i]:
      print("Mutated {} from {} to {}".format(i, p_dict[i], c_dict[i]))
      return True
  print("WARNING: no change found!!!")
  return False



def Mutate(cs, parent_model : Model) -> Model:
  model = copy.deepcopy(parent_model) 
  model.accuracy = None
    
  def op_mutation(model: Model) -> Model:
    operation = ""
    while "node" not in operation:
      operation = random.choice(model.arch().keys())
    model.exchange_one(operation)
    return model
  

  def cont_mutation( model : Model):
    operation = ""
    while ("node" in operation) or ("index" in operation) or operation == "": 
      operation = random.choice(model.arch().keys())
    model.perturb(operation)
    return model
    
    
     
  def hidden_state_mutation(model : Model) -> Model:
    operation = ""
    while "index" not in operation:
      operation = random.choice(model.arch().keys())
    model.exchange_one(operation)
    return model
    

  while not model_change(parent_model, model):
    model = random.choice([op_mutation, hidden_state_mutation])(model)    
  return model 


"""
def Mutate(cs, parent_model : Model) -> Model:
  model = copy.deepcopy(parent_model) 
  model.set_arch(csu.get_random_neighbor(model._arch, random.randint(1,999)))
  model_change(parent_model,model)
  return model
"""
def load_csv(file , cs):
  population = []
  history = []
  with open(file, "r") as csvfile:
    reader = csv.reader(csvfile)
    for row in reader:
      acc = float(row[0])
      rec = float(row[1])
      arch = eval(row[2])
      #if "channels" in arch:
      #  arch["channels"] = int(math.log(arch["channels"],2))
      #arch.pop("lr")
      #arch.pop("epochs")
      #arch.pop("p")
      model = Model(cs)
      model.set_arch(Configuration(cs,arch))
      model.accuracy = acc
      model.recall = rec
      population.append(model)
      history.append(model)
  return population, history
def regularized_evolution(configspace, worker , cycles, population_size, sample_size, sample_batch_size, load_file = None,load = False):
  """Algorithm for regularized evolution (i.e. aging evolution).

  Follows "Algorithm 1" in Real et al. "Regularized Evolution for Image
  Classifier Architecture Search".

  Args:
  cycles: the number of cycles the algorithm should run for.
  population_size: the number of individuals to keep in the population.
  sample_size: the number of individuals that should participate in each
      tournament.
  sample_batch_size: Number of children to generate before evaluating and adding to population
  Returns:
  history: a list of `Model` instances, representing all the models computed
      during the evolution experiment.
  """
  population = []
  history = []  # Not used by the algorithm, only used to report results.

  CS = configspace
  # Initialize the population with random models.
  if load_file ==None:
    train = train_eval( worker, sample_batch_size, "RegEvo.csv")
  elif not os.path.exists(load_file) or load == False:
    train = train_eval( worker, sample_batch_size, load_file)
  else:
    population , history = load_csv(load_file, CS)
    train = train_eval( worker, sample_batch_size, load_file )
    a = []
    r = []
    c= []
    print(history)
    for model in history:
      a.append(model.accuracy)
      r.append(model.recall)
      c.append(model.arch().get_dictionary())
    train.resume(a,r,c)
    if len(population) > population_size:
      population = population[-population_size:]
    for i in population:
      print(repr(i))
  while len(population) < population_size:
    model = Model( CS )
    model.set_arch( CS.sample_configuration() )
    population.append(model)
    history.append(model)
  train_and_eval_population(worker, population, sample_batch_size, train )
  # Carry out evolution in cycles. Each cycle produces a model and removes
  # another.
  children = []
  print_counter = 0
  while len(history) < cycles:

    print("Starting cycle", len(history) - population_size)
    # Sample randomly chosen models from the current population.

    if len(children) < sample_batch_size:
      sample = []
      while len(sample) < sample_size:
        # Inefficient, but written this way for clarity. In the case of neural
        # nets, the efficiency of this line is irrelevant because training neural
        # nets is the rate-determining step.
        candidate = random.choice(list(population))
        sample.append(candidate)

      # The parent is the best model in the sample.
      print("Selecting Parent via F1 Score")
      parent = max(sample, key=lambda i: F1(i.accuracy,i.recall))
      print("Scores of Parent F1: {} -- Acc: {} -- Recall: {}".format(F1(parent.accuracy,parent.recall),parent.accuracy,parent.recall))
      #parent = max(sample, key=lambda i: i.accuracy)

      # Create the child model and store it.
      child = Mutate(CS,parent)
      children.append(child)

    else:
      train_and_eval_population(worker, children, sample_batch_size, train)
      for i in children:
        population.append(i)
        history.append(i)
        population.pop(0)
      children = []
      # Remove the oldest model.
      best = max(population, key=lambda i: i.accuracy)
      print("--Current best model--")
      print("Accuracy:{} -- Recall: {} ".format(best.accuracy,best.recall))
      print("Architecture: ", best.arch())
      print("Population Size: ", len(population))
      print("Total Evaluations: ", len(history))
      print_counter +=1
      print_counter = 0
      accuracy_scores = []
      for i in history:
        accuracy_scores.append(i.accuracy)

      plot_scores(accuracy_scores)
  return history


def main(worker, configspace, load_file = "reg_evo.csv"):
  N = 5 #N best models to return
  pop_size = 500
  evaluations = 2500
  history = regularized_evolution(configspace, worker, cycles = evaluations, population_size =  pop_size, sample_size =25, sample_batch_size = 8, load_file = load_file, load = True)
  Architectures = []
  accuracy_scores = []
  recall_scores = []
  generations = list(range(evaluations))
  for i in history:
    accuracy_scores.append(i.accuracy)
    Architectures.append(i.arch)
    recall_scores.append(i.recall)
  top_acc = []
  top_rec = []
  top_config = []

  for i in range(N):
    indexs = accuracy_scores.index(max(accuracy_scores))
    print("Best accuracy: ", accuracy_scores[indexs])
    print("Best Hyperparameters: ", Architectures[indexs]())
    acc , rec , config = accuracy_scores[indexs] , recall_scores[indexs], Architectures[indexs]()
    top_acc.append(acc)
    top_rec.append(rec)
    top_config.append(config.get_dictionary())
    accuracy_scores.remove(max(accuracy_scores))
  return top_config,top_acc , top_rec



if __name__ == '__main__':
  main()



import random
import torch
import torch.nn as nn
import numpy as np
from torch import Tensor
from HPO.utils.model_constructor import Model
from torch.utils.data import Dataset, DataLoader, Sampler

def cal_acc(y,t):
  c = np.count_nonzero(y.T==t)
  tot = len(t)
  return c , tot

def convert_label_max_only(y):
  y = y.cpu().detach().numpy()
  idx = np.argmax(y,axis = 1)
  out = np.zeros_like(y)
  for count, i in enumerate(idx):
      out[count, i] = 1
  return idx

def stdio_print_training_data( iteration : int , outputs : Tensor, labels : Tensor , 
            epoch : int, epochs : int, correct :int , total : int , peak_acc : float , 
            loss : Tensor, n_iter, loss_list = None, binary = True):

  if binary == True:
    new_correct , new_total = cal_acc(outputs.ge(0.5).cpu().detach().numpy(), labels.cpu().detach().numpy())
  else:
    new_correct , new_total =  cal_acc(convert_label_max_only(outputs), convert_label_max_only(labels))
  
  correct += new_correct 
  total += new_total
  acc = correct / total 
  
  if acc > peak_acc:
    peak_acc = acc

  # Save the canvas
  print("Epoch (",str(epoch),"/",str(epochs), ") Accuracy: ","%.2f" % acc, 
              "Iteration(s) (", str(iteration),"/",str(n_iter), ") Loss: ",
              "%.2f" % loss," Correct / Total : {} / {} ".format(correct , total),  end = '\r')

  return correct ,total ,peak_acc



class BalancedBatchSampler(Sampler):
    def __init__(self, dataset,batch_size, indices = None):
        self.dataset = dataset
        if indices is None:
          self.indices = list(range(len(dataset)))
        else:
          self.indices = indices
        self.num_classes = dataset.get_n_classes()
        self.batch_size = batch_size
        # create a dictionary of class labels and their respective indices
        self.label_to_indices = {label: [] for label in range(self.num_classes)}
        for index in self.indices:
            label = dataset[index][1].item()
            self.label_to_indices[label].append(index)

        # ensure that batches are filled completely
        print( self.label_to_indices )
        #assert len(self.indices) % self.num_classes == 0
        self.samples_per_class = self.batch_size // self.num_classes
        self.remainder = self.batch_size % self.num_classes

    def __iter__(self):
        # shuffle labels for each class
        i = 0
        while i < len(self.indices) // self.batch_size:
          indices = []
          for label in self.label_to_indices:
              indices.extend(list(np.random.choice(self.label_to_indices[label],
                size = self.samples_per_class)))
          indices.extend(list(np.random.choice(self.label_to_indices[2],
                size = self.remainder)))    
          np.random.shuffle(indices)
          yield indices
          i+=1 

    def __len__(self):
        return len(self.indices) // self.batch_size


def highest_power_of_two(N):
    power = 1
    while N > 2**power:
      power+=1
    return 2**(power-1)




def collate_fn_padd(batch):
    '''
    Padds batch of variable length
    
    note: it converts things ToTensor manually here since the ToTensor transform
    assume it takes in images rather than arbitrary tensors.
    '''
    ## get sequence lengths
    lengths = torch.tensor([ t[0].shape[1] for t in batch ])    ## padd
    batch_samples = [ torch.transpose(t[0],0,1) for t in batch ]
    batch_samples = torch.nn.utils.rnn.pad_sequence(batch_samples ,batch_first = True)
    ## compute mask
    mask = (batch != 0)
    labels = []
    for pair in batch:
      one_hot_label = pair[1]
      labels.append(one_hot_label)
    labels = torch.stack(labels).squeeze()
    #labels = torch.Tensor([t[1] for t in batch])
    batch = torch.transpose(batch_samples , 1 , 2 )
    return batch, labels

def collate_fn_padd_x(batch):
    '''
    Padds batch of variable length

    note: it converts things ToTensor manually here since the ToTensor transform
    assume it takes in images rather than arbitrary tensors.
    '''
    ## get sequence lengths
    lengths = torch.tensor([ t.shape[1] for t in batch ])    ## padd
    batch_samples = [ torch.transpose(t,0,1) for t in batch ]
    batch_samples = torch.nn.utils.rnn.pad_sequence(batch_samples ,batch_first = True)
    ## compute mask
    mask = (batch != 0)
     

    batch = torch.transpose(batch_samples , 1 , 2 )
    return batch



"""Draw a live bar graph of frequency data (from Arduino FHT)."""

import collections
import curses
import math
import serial  # http://pyserial.sourceforge.net/
import time

_SERIAL_DEVICE = '/dev/tty.usbmodemfa141'
_GRAPH_MAX_Y = 100
_FAKE_DATA = True
_SMOOTHING = None if _FAKE_DATA else 10

_SERIAL_BAUD = 115200
_FHT_START_BYTE = 255


class CursesBarGraph:
  def __init__(self):
    self._window = None
    self._max = _GRAPH_MAX_Y

  def __enter__(self):
    self._window = curses.initscr()
    return self
  def set_max(self, val):
    self._max = val
  def __exit__(self, exc_type, exc_value, traceback):
    curses.endwin()

  def Update(self, values):
    assert self._window
    h, w = self._window.getmaxyx()
    per_bucket = max(1, math.ceil(float(len(values)) / (w - 1)))
    self._window.erase()
    for column_num, v in enumerate(_AveragedChunks(values, per_bucket)):
      assert column_num < w
      self._DrawBar(column_num, v, h)

    self._DrawAxisLabels(h, w, column_num, len(values))
    self._window.refresh()

  def _DrawBar(self, column_num, value, h):
    bar_len = max(0, min(h - 1, int(h * (value / self._max))))
    # vline draws from the starting coordinate towards positive y (down).
    self._window.vline((h - 1) - bar_len, column_num, ord('|'), bar_len)

  def _DrawAxisLabels(self, h, w, max_column, num_values):
    self._window.addstr(0, 0, str(self._max))
    self._window.addstr(h - 1, 0, str(0))
    max_column_str = str(num_values)
    self._window.addstr(
        h - 1,
        min(max_column, w - (len(max_column_str) + 1)),
        max_column_str)


def _AveragedChunks(iterable, n):
  summed_v = 0
  summed_count = 0
  for v in iterable:
    summed_v += v
    summed_count += 1
    if summed_count >= n:
      yield float(summed_v) / summed_count
      summed_v = 0
      summed_count = 0
  if summed_count > 0:
    yield float(summed_v) / summed_count


global _t
_t = 0.0
def MakeUpValues():
  time.sleep(0.05)
  global _t
  _t += .05
  values = []
  for i in range(200):
    values.append(_GRAPH_MAX_Y * (math.sin(_t + i/20.0) + 1) / 2.0)
  return values


def ReadValues(ser):
  values = []
  c = ord(ser.read())
  while c != _FHT_START_BYTE:
    values.append(c)
    c = ord(ser.read())
  return values


class Smoother:
  def __init__(self, window):
    self._n = window
    self._history = collections.defaultdict(lambda: list())

  def Smooth(self, values):
    smoothed = []
    for i, v in enumerate(values):
      history = self._history[i]
      history.append(v)
      if len(history) > self._n:
        history = history[-self._n:]
      smoothed.append(float(sum(history)) / self._n)
      self._history[i] = history
    return smoothed

def print_graph(values : list):
  smoother = Smoother(_SMOOTHING) if _SMOOTHING else None
  with CursesBarGraph() as bar_graph:
    try:
      values.append(values[-1]+1)
      bar_graph.set_max( max(values))
      if _SMOOTHING:
        values = smoother.Smooth(values)
      bar_graph.Update(values)
    except KeyboardInterrupt:
      pass



if __name__ == '__main__':
  print_graph([1,2,4,6,8,16,32])
  exit()




  values = [0]
  smoother = Smoother(_SMOOTHING) if _SMOOTHING else None
  with CursesBarGraph() as bar_graph:
    try:
      while True:
        values.append(values[-1]+1)
        bar_graph.set_max( max(values))
        if _SMOOTHING:
          values = smoother.Smooth(values)
        bar_graph.Update(values)
    except KeyboardInterrupt:
      pass



from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace
from HPO.algorithms.algorithm_utils import train_eval
import json
import networkx as nx 
import matplotlib.pyplot as plt
import random
from HPO.searchspaces.graph_search_space import Graph

def main(worker, configspace : ConfigurationSpace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]
  scores ,recall , pop = [], [], []
  history_scores = scores
  history_conf = pop
  last_max_indexs = None
  iteration = 0
  train = train_eval( worker , json_config)
  while True:
    configs = configspace.sample_configuration(SETTINGS["CORES"])
    model_1 = configs[0]
    model_2 = configs[1]
    #model_1 = convert_model(model_1,configspace.data)
    #model_2 = convert_model(model_2,configspace.data)
    configs.insert(0,join_models(model_1,model_2))
    scores ,recall , pop= train.eval(configs)
    history_conf.extend(pop)
    history_scores.extend(scores)

    top_indices = sorted(range(len(history_scores)), key=lambda i: history_scores[i], reverse=True)[:2]
    model_1 = history_conf[top_indices[0]]
    model_2 = history_conf[top_indices[1]]
    model_1 = convert_model(model_1,configspace.data)
    model_2 = convert_model(model_2,configspace.data)
    new_config = join_models(model_1,model_2)
    scores ,recall , pop = train.eval([new_config])
    history_conf.extend(pop)
    history_scores.extend(scores)


  
  best_score = max(scores)
  best_config = pop[scores.index(max(scores))]
  best_rec = recall[scores.index(max(scores))]


def plot_graph(edges):
    g= nx.DiGraph()
    g.add_edges_from(edges)
    plt.figure(figsize = (10,5))
    for i, layer in enumerate(nx.topological_generations(g)):
        for n in layer:
            g.nodes[n]["layer"] = i
    pos = nx.multipartite_layout(g,subset_key="layer", align="vertical")
    for i in pos:
        temp= pos[i]
        temp[1] += random.uniform(-0.3,0.3)

    nx.draw(
         g, edge_color='black',pos = pos , width=3, linewidths=5,
         node_size=2000, node_color='pink', alpha=1,font_size = 20
         )

    plt.axis('off')
    plt.savefig("test") 
    plt.show()
    return g


def convert_model(m,data):
  edges = m["graph"]
  ops = m["ops"]
  g = nx.DiGraph()
  g.add_edges_from(edges)
  model = Graph(edges,g.nodes,ops,data)
  return model

def join_models(model_1,model_2):

  #GET MAX VALUE IN MODEL_1 
  max_val = model_1.max_vertex + 1
  #INCREASE ALL VALUES IN MODEL_2 BY MAX VALUE
  model_2.rename_vertices(max_val)
  #COMBINE
  m_dict_1 = model_1()
  m_dict_2 = model_2()
  print("graph 1: ",m_dict_1["graph"])
  print("graph 2: ",m_dict_2["graph"])
  m_dict_1["graph"].extend(m_dict_2["graph"])
  print("graph joined: ",m_dict_1["graph"])
  m_dict_1["ops"].update(m_dict_2["ops"])
  plot_graph( m_dict_1["graph"])
  return m_dict_1




import matplotlib.pyplot as plt
import torch
import random
from torch.nn.functional import interpolate
import time
import numpy as np
from operator import itemgetter
from scipy.interpolate import CubicSpline


def initialise_augmentations(augmentation_data)->list:
  if augmentation_data:
    augs = []
    for i in augmentation_data:
      augs.append(eval(i.split("_")[0])(**augmentation_data[i]) )
    return augs
  else:
    return False

class Augmentation(object):
  def __init__(self,rate,device):
    self.device = device
    self.rate = rate
  def __call__(self,x,y):
    rate = self.rate
    while random.random() < rate:
      x,y = self.call(x,y)
      rate -= 1
    else:
      return x,y
  def call(self,x,y):
    print("called parent class function :(")

class Jitter(Augmentation):
  def __init__(self,sigma = 0.05,rate = 0.3,device = None):
    super(Jitter,self).__init__(rate,device)
    self.n = torch.distributions.normal.Normal(loc=torch.FloatTensor([0]).cuda(device =device), scale=torch.FloatTensor([sigma]).cuda(device = device))
    self.__name__ = "jitter"
  def call(self,x,y):
    return torch.add(x,self.n.sample(x.shape).squeeze()),y 

class Scaling(Augmentation):
  def __init__(self,sigma = 0.05,rate = 0.3 ,device = None):
    super(Scaling,self).__init__(rate,device)
    self.n = torch.distributions.normal.Normal(loc=torch.FloatTensor([1]).cuda(device =device), scale=torch.FloatTensor([sigma]).cuda(device = device))
    self.__name__ = "scaling"
  def call(self,x,y):
    s = self.n.sample((x.shape[0],x.shape[1])).squeeze()
    return torch.mul(x, s),y

class Crop(Augmentation):
  def __init__(self, crop_min = 0.5,rate = 0.3, crop_max = 0.99, device = None):
    super(Crop,self).__init__(rate,device)
    self.__name__ = "crop"
    self.crop_min = crop_min
    self.crop_max = crop_max
  def call(self,x,y):
    sig_len = x.shape[1]
    length= random.uniform(self.crop_min,self.crop_max)
    length = int(length * sig_len)
    start = random.randint(0,(sig_len - length))
    return x[:,start:(length+start)],y

    if random.choice([0,1]) == 1:
      return  x[:,:length],y
    else:
      return  x[:,(sig_len-length):],y

class WindowWarp(Augmentation):
  def __init__(self, num_warps = 3, ratios = [0.5,0.75,2],size = 0.3,rate = 0.3, device = None):
    super(WindowWarp,self).__init__(rate,device)
    self.__name__ = "window_warp"
    self.ratios = ratios
    self.num_warps = num_warps
    self.size = size
  def call(self,x,y):
    for i in range(self.num_warps):
        start = random.randint(1, x.shape[1]-10) 
        end = min([x.shape[1],start+random.randint(2, int(x.shape[1]*self.size))])
        out= interpolate(x[:,start:end].unsqueeze(0), scale_factor = random.choice(self.ratios)).cuda(device = self.device).squeeze(0)
        x = torch.cat((x[:,:start],out,x[:,end:]),dim = 1)
    if x.shape[1] > 20000:
        print("Window Warp exceeded size: {}".format(x.shape[1]))
    return x,y
class CutOut(Augmentation):
  def __init__(self,perc=.1,rate = 0.3,device = None):
    super(CutOut,self).__init__(rate,device)
    self.perc = perc
    self.__name__ = "cut_out"
  def call(self,x,y):
    seq_len = x.shape[1]    
    win_len = int(self.perc * seq_len)    
    start = np.random.randint(0, seq_len-win_len-1)    
    end = start + win_len    
    start = max(0, start)    
    end = min(end, seq_len)    
    x[:,start:end] = 0    
    return x,y 

class MixUp(Augmentation):
  def __init__(self, m = 0.3 , rate = 0.3 , device = None):
    super(MixUp,self).__init__(rate,device)
    self.dist = torch.distributions.beta.Beta(m,m)
    self.mix_sample = None
    self.mix_label = None 
  def __call__(self,x,y):
    if random.random() < self.rate:
      if self.mix_sample == None:
        self.mix_sample = x
        self.mix_label = y
        return x,y
      elif self.mix_sample.shape[1] != x.shape[1]:
        print("switching to val size")
        self.mix_sample = x
        self.mix_label = y
        return x,y
      else: 
        return self.call(x,y)
    else:
      return x,y

  def call(self,x,y):
    DEBUG = False
    #x = [batch, channels , length]
    #y = [batch, 1]
    mix = self.dist.sample()
    MTS_1 = x.clone()
    MTS_2 = self.mix_sample.clone()
    LAB_1 = y.clone()
    LAB_2 = self.mix_label.clone()
    x = (MTS_1 * mix)  + (MTS_2 *(1-mix))
    y = (LAB_1 * mix) + (LAB_2 * (1-mix))
    if DEBUG == True:
      print("mix value is {} and label value is {}".format(mix,y))
      plt.plot(MTS_1[26,:].cpu(),label = "series 1",alpha = 0.7)
      plt.plot(MTS_2[26,:].cpu(),label = "series 2",alpha = 0.7)
      plt.plot(x[26,:].cpu(),label = "Mix",alpha = 0.7)
      plt.legend()
      plt.show()
    return x ,y

def mix_up(x,y, m = 0.3, device = None):
    dist = torch.distributions.beta.Beta(m,m)
    DEBUG = False
    #x = [batch, channels , length]
    #y = [batch, 1]
    for i in range(x.shape[0]):
      i_2 = list(range(x.shape[0]))
      i_2.remove(i)
      i_2 = random.choice(i_2)
      mix = dist.sample()
      MTS_1 = x[i,:,:].clone()
      MTS_2 = x[i_2,:,:].clone()
      LAB_1 = y[i]
      LAB_2 = y[i_2]
      x[i,:,:] = (MTS_1 * mix)  + (MTS_2 *(1-mix))
      y[i] = (LAB_1 * mix) + (LAB_2 * (1-mix))
      if DEBUG == True:
        print("mix value is {} and label value is {}".format(mix,y[i]))
        plt.plot(MTS_1[26,:].cpu(),label = "series 1",alpha = 0.7)
        plt.plot(MTS_2[26,:].cpu(),label = "series 2",alpha = 0.7)
        plt.plot(x[i,26,:].cpu(),label = "Mix",alpha = 0.7)
        plt.legend()
        plt.show()
    return x ,y

def cut_mix(x,y, perc=.1, device = None):    
    seq_len = x.shape[2]    
    win_len = int(perc * seq_len)    
    start = np.random.randint(0, seq_len-win_len-1)    
    end = start + win_len    
    for i in range(x.shape[0]):    
        start = max(0, start)    
        end = min(end, seq_len)    
        x[i,:,start:end] = x[random.randint(0,x.shape[0]-1),:,start:end]    
    # return new_ts, ts[start:end, ...]    
    return x,y 






















def rotation(x : torch.Tensor,y):
    flip = torch.randint(0,2,size = (x.shape[0],x.shape[2]))
    flip = torch.where(flip != 0, flip, -1)
    rotate_axis = torch.arange(x.shape[2])
    s = np.random.shuffle(np.array(range(x.shape[2])))
    rotate_axis[np.array(range(x.shape[2]))] = rotate_axis[s].clone()
    return flip[:,None,:] * x[:,:,rotate_axis],y



def permutation(x : torch.Tensor,y, max_segments=5, seg_mode="equal"):
    #make array of ints from 1 to window_length
    num_seqments_per_batch = torch.randint(0 , max_segments, 
        size = (x.shape[0],))
    #loop through batches and split into the 
    #defined number of segments for each 
    for idx, sequence in enumerate(x):
        if num_seqments_per_batch[idx] > 1:

            #returns a list of tensors
            chunk_list = torch.chunk(sequence, num_seqments_per_batch[idx], dim = 0)
            #Shuffle
            shuffled_chunks = itemgetter(*np.random.permutation(len(chunk_list)))(chunk_list)
            permutated_sequence = torch.cat(shuffled_chunks, dim = 0)
            x[idx] = permutated_sequence


    return x,y






def magnitude_warp(x : torch.Tensor,y, sigma=0.2, knot=4):
    def h_poly(t):
        tt = t[None, :]**torch.arange(4, device=t.device)[:, None]
        A = torch.tensor([
            [1, 0, -3, 2],
            [0, 1, -2, 1],
            [0, 0, 3, -2],
            [0, 0, -1, 1]
        ], dtype=t.dtype, device=t.device)
        return A @ tt


    def interp(x, y, xs):
        m = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
        m = torch.cat([m[[0]], (m[1:] + m[:-1]) / 2, m[[-1]]])
        idxs = torch.searchsorted(x[1:], xs)
        dx = (x[idxs + 1] - x[idxs])
        hh = h_poly((xs - x[idxs]) / dx)
        return hh[0] * y[idxs] + hh[1] * m[idxs] * dx + hh[2] * y[idxs + 1] + hh[3] * m[idxs + 1] * dx

    orig_steps= torch.linspace(0, x.shape[1]-1, x.shape[1])
    n = torch.distributions.normal.Normal(loc=1., scale=sigma)
    random_warps = n.sample((x.shape[0],knot+2 , x.shape[2]))
    warp_steps = (torch.ones((x.shape[2],1))*(torch.linspace(0, x.shape[1]-1., knot+2))).T
    ret = torch.zeros(x.shape)
    warper = torch.zeros(orig_steps.shape[0],x.shape[2])
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            warper[:,dim] =  interp(warp_steps[:,dim], random_warps[i,:,dim],  orig_steps)

        ret[i] = pat * warper

    return ret,y

if __name__ == "__main__":

    import torch.nn as nn
    from torch import Tensor
    from HPO.data.teps_datasets import Train_TEPS
    from torch.utils.data import DataLoader
    import random
    import timeit
    import matplotlib.pyplot as plt
    jitter = Jitter(device = 0)
    scaling = Scaling(device = 0)
    window_warp = WindowWarp(device = 0)
    crop = Crop(device = 0)
    cut_out = CutOut(device = 0)
    funcs = [crop,jitter,scaling,window_warp,cut_out]
    batch_size = 1
    window_length = 500
    features = 27
    train_dataset = Train_TEPS()
    x,y = next(iter(train_dataset))
    print(x.shape)
    train_dataloader = DataLoader( train_dataset, batch_size=batch_size,
      shuffle = True,drop_last=True)
    for i,(s,l) in enumerate(train_dataloader):
        x = s.cuda().squeeze()
        y = l.cuda()
        print("Shape of sample: {}".format(x.shape))
        if i > 0:
          break
        for func in funcs:
            """
            print(x.shape)
            print(func)
            #plt.plot(x[0,26,:].cpu(),label = "orig")
            x_p,y_p = func(x,y,device = 0)
            x_p = x_p[0,26,:].cpu()
            #plt.plot(x_p ,alpha = 0.5,label = "aug")
            #plt.legend()
            #plt.show()
            """
            print("from __main__ import {}, {}, {}".format(func.__name__, "x","y"))
            t = timeit.timeit("{}(x,y)".format(func.__name__), "from __main__ import {}, {}, {}".format(func.__name__, "x","y"), number = 1000)
            print("Total time for ",func.__name__,": ",t , " Seconds")
            print("Time per sample: {}".format(t/batch_size))




from HPO.searchspaces.utils import *
import networkx as nx
import random

def build_cell_graph(n_nodes = 4):
    graph = []
    ops = dict()
    source_list = [1,2]
    for end_node in range(3,n_nodes+3):
        for source_node in source_list:
            graph.append((source_node,end_node))
            ops["{}_combine".format(end_node)] = "ADD"  
        source_list.append(end_node)
        graph.append((end_node,n_nodes +3))
    ops["{}_combine".format(n_nodes +3 )] = "CONCAT" 
    graph = prune_cell(graph,n_nodes)
    graph = fix_concat_inputs(graph,n_nodes)
    return graph, ops

def fix_concat_inputs(graph,n_nodes):
    nodes = [i for i in range(3,n_nodes+3)]
    output = max(nodes) + 1 
    for edge in graph:
        if edge[0] in nodes and edge[1] in nodes and (edge[0],output) in graph:

            graph.remove((edge[0],output))

    print(graph)
    return graph


def generate_cell_ops(graph,ops,data,n_nodes = 4):
    nodes = [i for i in range(3,n_nodes+3)]
    op_edges = []
    skip_edges = []
    for edge in graph:
        if edge[1] in nodes:
            op_edges.append(edge)
        else:
            skip_edges.append(edge)

    g = nx.DiGraph()
    g.add_edges_from(op_edges)
    ops_temp = generate_op_names(g)
    ops_temp = random_ops_unweighted(ops_temp, data)
    ops_temp = random_activation_unweighted(ops_temp,data)
    ops_temp = random_normalisation_unweighted(ops_temp,data)
    for i in ops_temp:
        if not i in ops:
            ops[i] = ops_temp[i]
    g = nx.DiGraph()
    g.add_edges_from(skip_edges)
    ops_temp = generate_skip(g)
    for i in ops_temp:
        if not i in ops:
            ops[i] = ops_temp[i]

    return ops


def prune_cell(graph, n_nodes = 4):
    nodes = [i for i in range(3,n_nodes+2)]
    count_list = {}
    for i in nodes:
        count_list[i] = 0
    for edge in graph:
        if edge[1] in nodes:
            count_list[edge[1]] += 1
    while max(count_list.values()) > 2:
        edge = random.choice(graph)
        if edge[1] in nodes and count_list[edge[1]] > 2:
            graph.remove(edge)
            count_list[edge[1]] -= 1
    return graph




def replace_cell_inputs(graph,ops,outputs):
    g = nx.DiGraph()
    g.add_edges_from(graph)
    graph = nx.relabel_nodes(g, {"K-1": outputs[-1],"K-2": outputs[-2]}).edges
    for i in ops:
        if "K-1" in i:
            ops[i.replace("K-1",str(outputs[-1]))] = ops.pop(i)
        if "K-2" in i:
            ops[i.replace("K-2",str(outputs[-2]))] = ops.pop(i)
    return list(graph), ops


def build_macro( n_nodes = 8, n_cells = 5, reduction_freq = 3):
    """
    Builds the graph of a cell style search space
    """
    graph = []
    ops = {}
    cell_outputs = ["S","S"]
    for layer in range(n_cells):
        graph_new , ops_new = build_cell_graph(n_nodes)


        
        graph  = graph_joiner(graph,graph_new)
        ops  = op_joiner(ops,ops_new)
        #graph , ops = replace_cell_inputs(graph,ops,cell_outputs) #REPLACE "K-1","K-2" PLACEHOLDERS
        cell_outputs.append((n_nodes+3) + (n_nodes+3)*layer)
        graph.extend([(cell_outputs[-3],cell_outputs[-1]-((n_nodes+2))),((cell_outputs[-2],cell_outputs[-1]-(n_nodes+1)))])
        if layer % reduction_freq == 0 and layer != 0 and layer != n_cells -1 :
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+2))] = 2
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+1))] = 2
            ops["{}_stride".format(cell_outputs[-1]-(n_nodes+2))] = 2
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+1))] = 2


    g = nx.DiGraph()
    g.add_edges_from(graph)
    nodes = list(nx.topological_sort(g))
    for idx ,(e_0,e_1) in enumerate(graph):
        if e_1 == nodes[-1]:
            hold = (list(graph[idx])[0], "T")
            graph[idx] = tuple(hold)

    hold_dict = {}
    rm_list = []
    for idx ,op in enumerate(ops):
        if op.split("_")[0] == nodes[0]:
            ops[("{}_{}".format("S",op.split("_")[1]))] = ops[op]
        if int(op.split("_")[0]) == nodes[-1]:
            hold_dict[("{}_{}".format("T",op.split("_")[1]))] = ops[op]
            rm_list.append(op)
    for i in rm_list:
        del ops[i]
    ops.update(hold_dict)
    return graph, ops

def build_macro_repeat(data, n_nodes = 8, n_cells = 5, reduction_freq = 3,stride= 2, channel_ratio =2 ):
    """
    Builds the graph of a cell style search space
    """
    graph = []
    ops = {}
    cell_outputs = ["S","S"]
    graph_new , ops_new = build_cell_graph(n_nodes)
    ops_new = generate_cell_ops(graph_new,ops_new,data,n_nodes = n_nodes)
    for layer in range(n_cells):



        
        graph  = graph_joiner(graph,graph_new)
        ops  = op_joiner(ops,ops_new)
        #graph , ops = replace_cell_inputs(graph,ops,cell_outputs) #REPLACE "K-1","K-2" PLACEHOLDERS
        cell_outputs.append((n_nodes+3) + (n_nodes+3)*layer)
        graph.extend([(cell_outputs[-3],cell_outputs[-1]-((n_nodes+2))),((cell_outputs[-2],cell_outputs[-1]-(n_nodes+1)))])
        if layer % reduction_freq == 0 and layer != 0 and layer != n_cells -1 :
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+2))] = channel_ratio
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+1))] = channel_ratio
            ops["{}_stride".format(cell_outputs[-1]-(n_nodes+2))] = stride
            ops["{}_channel_ratio".format(cell_outputs[-1]-(n_nodes+1))] = channel_ratio


    g = nx.DiGraph()
    g.add_edges_from(graph)
    nodes = list(nx.topological_sort(g))
    for idx ,(e_0,e_1) in enumerate(graph):
        if e_1 == nodes[-1]:
            hold = (list(graph[idx])[0], "T")
            graph[idx] = tuple(hold)

    hold_dict = {}
    rm_list = []
    for idx ,op in enumerate(ops):
        if op.split("_")[0] == nodes[0]:
            ops[("{}_{}".format("S",op.split("_")[1]))] = ops[op]
        if int(op.split("_")[0]) == nodes[-1]:
            hold_dict[("{}_{}".format("T",op.split("_")[1]))] = ops[op]
            rm_list.append(op)
    for i in rm_list:
        del ops[i]
    ops.update(hold_dict)
    return graph, ops
            

        


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import random
    cells = 3 
    c = build_macro()
    print(c[1])
    def plot_graph(edges):
        g= nx.DiGraph()
        g.add_edges_from(edges)
        plt.figure(figsize = (50,10))
        for i, layer in enumerate(nx.topological_generations(g)):
            for n in layer:
                g.nodes[n]["layer"] = i
        pos = nx.multipartite_layout(g,subset_key="layer", align="vertical")
        for i in pos:
            temp= pos[i]
            temp[1] += random.uniform(-0.3,0.3)
    
        nx.draw(
             g, edge_color='black',pos = pos , width=1, linewidths=5,
             node_size=2000, node_color='pink', alpha=0.9,font_size = 35,with_labels=True
             )
        plt.axis('off')
        plt.savefig("test") 
    plot_graph(c[0])
    print(c)


from HPO.utils.files import load_obj
import os
import torch
import numpy as np
from HPO.utils.model_constructor import Model
from HPO.data.datasets import repsol_unlabeled
from torch.utils.data import DataLoader
import matplotlib.pyplot as plt
from HPO.utils.worker_helper import collate_fn_padd_x
def load_model( model_code ):
	hyperparameter = load_obj( "/home/snaags/scripts/HPO/src/HPO/model_zoo/hps/{}".format(model_code) )
	model = Model((27, ) , 2 , hyperparameters = hyperparameter)
	model.load_state_dict(torch.load("/home/snaags/scripts/HPO/src/HPO/model_zoo/{}".format(model_code)))

	return model 

def generate_soft_labels( model ):
	batch_size = 1
	dataset = repsol_unlabeled()
	dataloader = DataLoader( dataset,
		shuffle = False,batch_size = 1,drop_last=False,pin_memory=True)
	output_list = []
	pred_list = []
	input_list = []
	model = model.cuda()
	with torch.no_grad(): #disable back prop to test the model
		model = model.eval()
		for i, (inputs) in enumerate( dataloader):

			inputs = inputs.cuda(non_blocking=True)
			outputs = model(inputs.float())

			preds = torch.argmax(outputs.view( batch_size ,2),1).long().cpu().numpy()

			output_list.append(outputs.cpu().numpy())
			input_list.append(inputs.cpu().numpy())
			pred_list.append(preds)

		y = np.asarray(output_list)
		input_list = np.asarray(input_list).reshape(27,-1)
		print(y.shape)
	dataset.add_labels(y)
	return dataset

if __name__ == "__main__":
	models = os.listdir("/home/snaags/scripts/HPO/src/HPO/model_zoo/")
	models = sorted(models)
	print(models[-2])
	model = load_model(models[-2])
	generate_soft_labels( model )



from HPO.searchspaces.resnet import build_resnet, ResNet50
from HPO.searchspaces.graph_search_space import GraphConfigSpace
from HPO.searchspaces.utils import *
from HPO.searchspaces.cell import build_macro, build_macro_repeat
import random
#ResNet
#Cell
#Hierarchical
#Graph



class FixedModel:
  """
  Fixed model can be used for testing or hyperparameter search etC
  """
  def __init__(self, model):
    self.model = model
  def sample_configuration(self, num):
    return [self.model] * num


class ResNetSearchSpace:
  def __init__(self, JSON):
    self.data = JSON["ARCHITECTURE_CONFIG"]
    self.EDGE_OPERATIONS = self.data["EDGE_OPERATIONS"]


  def generate_layers(self):
    x = [1]
    for i in range(random.randint(self.data["MIN_LAYERS"]-1,self.data["MAX_LAYERS"]-1)):
      if random.random() < (1/len(x)):
        x.append(1)
      else:
        x[random.randint(0,len(x)-1)] += 1
    return x
      


  def sample_configuration(self, num):
    configs = []
    for i in range(num):
      m = build_resnet(self.generate_layers())
      m["ops"] = random_ops_unweighted(m["ops"],self.data)
      m["ops"] = random_activation_unweighted(m["ops"],self.data)
      m["ops"] = random_normalisation_unweighted(m["ops"],self.data)
      configs.append(m)
    print(configs)
    return configs


def resnet_search_space(JSON_CONFIG):
  return ResNetSearchSpace(JSON_CONFIG)

def resnet50(JSON_CONFIG):
    return FixedModel(ResNet50())
     

class FixedTopology:
  def __init__(self, JSON):
    with open(JSON) as conf:
      data = json.load(conf)
    self.model = model

  def   generate_ops(self):
    ops = self.model["ops"]
    "OP"



def generate_ops(op_set, location , name ,parameter_list = None):
  if parameter_list == None:
    parameter_list =[]
  if len(op_set) == 1:
    for i in location:
      parameter_list.append(CSH.Constant("{}_{}".format(i,name),value = op_set[0]))
  else:
    for i in location:
      parameter_list.append(CSH.CategoricalHyperparameter('{}_{}'.format(i,name), choices=op_set))
  return parameter_list

class SearchSpace:
  def __init__(self,JSON):
    #LOAD JSON DATA
    GRAPH_SIZE = None
    GRAPH_RATE = None
    EDGE_OPERATIONS = None 
    
    #GENERATE GRAPH
    self.edges_op = []
    self.nodes_op = []
    self.graph = []
    self.hyperparameters = []

    #GENERATE EDGE OPS
    self.hyperparameters = generate_ops(EDGE_OPERATIONS, self.graph, "OP",self.hyperparameters)
    
    #GENERATE NODE OPS
    self.hyperparameters = generate_ops(NODE_ACTIVATIONS, self.nodes, "activation",self.hyperparameters)
    self.hyperparameters = generate_ops(NODE_NORMALISATION, self.nodes, "normalisation",self.hyperparameters)
    
    #Here the same node can be selected twice to allow for a larger downsampling
    #THIS NEEDS TO BE CHANGED!!!!
    self.hyperparameters = generate_ops(self.nodes, downsample_quantity, "downsample",self.hyperparameters)

    #This channel variation should be for that op only allowing for bottlenecks and expansions.
    self.hyperparameters = generate_ops([0.25,0.5,1,2,4], self.nodes, "channel_ratio",self.hyperparameters)
   

def graph_config(JSON):
  print("here")
  graph = GraphConfigSpace(JSON)
  return graph

class CellSpace:
  def __init__(self,JSON):
    self.data = JSON["ARCHITECTURE_CONFIG"]
    self.n_nodes = 4
    self.cells = 5
    self.reduce = 1

  def sample_configuration(self,n):
    configs = []
    for i in range(n):
      self.g = nx.DiGraph()
      model = build_macro( n_nodes = self.n_nodes, n_cells = self.cells, reduction_freq = self.reduce)
      self.g.add_edges_from(model[0])
      ops = generate_op_names(self.g)
      ops = random_ops_unweighted(ops, self.data)
      ops = random_activation_unweighted(ops,self.data)
      ops = random_normalisation_unweighted(ops,self.data)
      #ops = random_combine_unweighted(ops,self.data)
      #ops = random_strides(ops,self.data["STRIDE_COUNT"])
      ops.update(model[1])
      del ops["T_stride"]
      del ops["T_channel_ratio"]
      del ops["S_stride"]
      del ops["S_channel_ratio"]
      configs.append({"graph": model[0], "ops": ops})
    return configs

class CellSpaceRepeat:
  def __init__(self,JSON):
    self.data = JSON["ARCHITECTURE_CONFIG"]
    self.n_nodes = 4
    self.cells = self.data["N_CELLS"]
    self.reduce = 1

  def sample_configuration(self,n):
    configs = []
    for i in range(n):
      self.g = nx.DiGraph()
      model_stride = random.choice(self.data["STRIDE_RATE"])
      model_stride_channel_ratio = random.choice(self.data["CHANNEL_DEPTH_RATE"])
      model = build_macro_repeat( data = self.data, n_nodes = self.n_nodes, n_cells = self.cells, reduction_freq = self.reduce,stride = model_stride,channel_ratio = model_stride_channel_ratio)
      self.g.add_edges_from(model[0])
      #ops = generate_op_names(self.g)
      #ops = random_ops_unweighted(ops, self.data)
      #ops = random_activation_unweighted(ops,self.data)
      #ops = random_normalisation_unweighted(ops,self.data)
      #ops = random_combine_unweighted(ops,self.data)
      #ops = random_strides(ops,self.data["STRIDE_COUNT"])



       
      ops_temp = generate_skip(self.g)
      for i in ops_temp:
          if not i in model[1]:
              model[1][i] = ops_temp[i]
      #ops.update(model[1])
      del model[1]["T_stride"]
      del model[1]["T_channel_ratio"]
      del model[1]["T_channel"]
      del model[1]["S_stride"]
      del model[1]["S_channel_ratio"]

      for i in model[1]:
        if "combine" in i:
          if model[1][i] == 1:
            model[1][i] = "ADD"
      model[1]["s_rate"] = model_stride
      model[1]["s_c_ratio"] = model_stride_channel_ratio
      configs.append({"graph": model[0], "ops": model[1]})
    return configs

def cell_config(JSON):
  return CellSpace(JSON)

def cell_repeat_config(JSON):
  return CellSpaceRepeat(JSON)




import math
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.genotypes import NASNet
from HPO.utils.DARTS_utils import config_space_2_DARTS
import torch
import numpy as np
import seaborn as sns
from HPO.searchspaces import DARTS_config
import csv
hps_best = {'T_0': 3,
 'T_mult': 1,
 'batch_size': 2,
 'channels': 19,
 'epochs': 41,
 'layers': 3,
 'lr': 0.002993825743228492,
 'normal_index_0_0': 0,
 'normal_index_0_1': 1,
 'normal_index_1_0': 1,
 'normal_index_1_1': 1,
 'normal_index_2_0': 3,
 'normal_index_2_1': 3,
 'normal_index_3_0': 4,
 'normal_index_3_1': 1,
 'normal_node_0_0': 'skip_connect',
 'normal_node_0_1': 'dil_conv_3x3',
 'normal_node_1_0': 'sep_conv_3x3',
 'normal_node_1_1': 'dil_conv_5x5',
 'normal_node_2_0': 'max_pool_3x3',
 'normal_node_2_1': 'sep_conv_3x3',
 'normal_node_3_0': 'dil_conv_3x3',
 'normal_node_3_1': 'skip_connect',
 'p': 0.017718387446598843,
 'reduction_index_0_0': 0,
 'reduction_index_0_1': 1,
 'reduction_index_1_0': 0,
 'reduction_index_1_1': 1,
 'reduction_index_2_0': 3,
 'reduction_index_2_1': 1,
 'reduction_index_3_0': 3,
 'reduction_index_3_1': 4,
 'reduction_node_0_0': 'max_pool_3x3',
 'reduction_node_0_1': 'avg_pool_3x3',
 'reduction_node_1_0': 'skip_connect',
 'reduction_node_1_1': 'skip_connect',
 'reduction_node_2_0': 'max_pool_3x3',
 'reduction_node_2_1': 'max_pool_3x3',
 'reduction_node_3_0': 'sep_conv_3x3',
 'reduction_node_3_1': 'avg_pool_3x3'}

configspace = DARTS_config.init_config()

def load_csv(file):

  acc_l = []
  rec_l = []
  config_l = []
  with open(file, "r") as csvfile:
    reader = csv.reader(csvfile)
    for row in reader:
      acc = float(row[0])
      rec = float(row[1])
      config = eval(row[2])
      acc_l.append(acc)
      rec_l.append(rec)
      config_l.append(config)

  return acc_l, rec_l , config_l

C = list()

def hamming_distance(c_1, c_2):
  
  x = torch.cdist(c_1,c_2, p = 0)
  
  return x

def hamming_kernel(C_list : list):
  K_h = torch.zeros(len(C_list), len(C_list))
  for i, c_i in enumerate(C_list):
    for j,c_j in enumerate(C_list):
      K_h[i,j] = len(c_j[0]) - hamming_distance(c_i,c_j)
  return K_h

def score(K_h):
  x = np.linalg.slogdet(K_h.detach().numpy())
  return x[1]



def activation_hook(inst, inp , out):
  hold = torch.reshape(out, (1, -1))
  C.append(hold)  

import random
acc, rec, conf = load_csv("/home/snaags/uist/RegEvo.csv")
scores = {}
for _, (c,a_acc) in enumerate(zip(conf,acc)):
  if random.random() < 0.9:
    continue
  print("Model: {}".format(_))
  gen = config_space_2_DARTS(c)
  model = NetworkMain(16,c["channels"],2,3,False, 0,gen) ##Initialise Network
  
  ##Attach Activation Hooks

  x = repr(model)
  for i in model.cells:
    if "FactorizedReduce" not in repr(i.preprocess0) :
      for x in i.preprocess0.op:
        if repr(x) == "ReLU()":
          x.register_forward_hook(activation_hook)
      for x in i.preprocess1.op:
        if repr(x) == "ReLU()":
          x.register_forward_hook(activation_hook)
    
    #print("Activation Functions: {} {}".format(act1, act2))
    for j in i._ops:
      if "Conv" in repr(j) and "FactorizedReduce" not in repr(j):
        for j_i in j.op:
          if repr(j_i ) == "ReLU()":
            j_i.register_forward_hook(activation_hook)
  
      
  
  from sklearn.preprocessing import StandardScaler
  
  
  
  ##Generate Activation Patterns
  
  N = 64
  c_list = list()
  for __ in range(N):
  
    C = list()
    batch = torch.rand(1,16,8)
    model(batch)
    c_out = torch.cat(C,dim = 1 )
    c_map = c_out.clone()
    c_out[c_map != 0] = 1
    c_list.append(c_out)
  K_h = hamming_kernel(c_list)
  SS = StandardScaler()
  K_h_out = SS.fit_transform(K_h.detach().numpy())
  scores[_] = score(K_h)
  print("Final Score: {} -- acc : {}".format(score(K_h), a_acc))

##Match
acc_out = []
score_out = []
for i in scores:
  score_out.append( scores[i])
  acc_out.append( acc[i])
plt.scatter(acc_out , score_out)
plt.show()


    
      



import numpy as np 
import time
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model, auto_train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute( ID = None, configs=None , gpus=None , res = None  , config = None):
  device = None
  print("Starting process: {}".format(ID))
  while not configs.empty():
    try:
      if device == None:
        device = gpus.get(timeout = 10)
      config = configs.get(timeout = 10)
    except Empty:
      if device != None:
        gpus.put(device)
      
    except:
      torch.cuda.empty_cache()
      if device != None:
        gpus.put(device)
      return
    if config != None:
      print("Got Configuration!")

    if device != None:
      print("Starting config with device: {}".format(device))
      complete = False
      crashes = 0
      acc , rec =  _compute(hyperparameter = config , cuda_device = device)
      while not complete:
        try:
          
          complete = True
        except:
          crashes +=1
          print("Model crash: {} ".format(crashes))
          time.sleep(60)
          if crashes == 2:
            print("Final Crash giving score of zero")
            acc , rec = 0 , 0 
            complete = True
      res.put([config , acc , rec ]) 

  torch.cuda.empty_cache()
   
def _compute(hyperparameter,budget = 1, in_model = None , train_path = None,  test_path = None, cuda_device = None,plot_queue = None, model_id = None, binary = True):
  ### Configuration 
  THRESHOLD = 0.4 #Cut off for classification
  batch_size = 64
  if cuda_device == None:
     cuda_device = 1# torch.cuda.current_device()
  
  ##Set up augmentations##
  jitter = aug.Jitter(device = cuda_device,sigma = 0.125, rate = 0.5)
  crop = aug.Crop(device = cuda_device, rate = 0.8, crop_min = 0.3 , crop_max = 0.98)
  scaling = aug.Scaling(device = cuda_device)
  window_warp = aug.WindowWarp(device = cuda_device,rate = 0.5)
  cut_out = aug.CutOut(device = cuda_device)
  mix_up = aug.MixUp(device = cuda_device,m = 0.2, rate = 0.3)
  augmentations = [jitter,crop,scaling, window_warp, cut_out]

  dataset_train = Train_TEPS(augmentations= augmentations,samples_per_class = 500,device = cuda_device,one_hot = False,binary = False)

  #dataset_test_full = Test_TEPS()
  torch.cuda.set_device(cuda_device)

  print("Cuda Device Value: ", cuda_device)
  gen = config_space_2_DARTS(hyperparameter,reduction = True)
  print(gen)


  n_classes = dataset_train.get_n_classes()
  multibatch = False
  torch.cuda.empty_cache()
  trainloader = torch.utils.data.DataLoader(
                      dataset_train,collate_fn = collate_fn_padd,shuffle = True,
                      batch_size=batch_size, drop_last = True)


  model = NetworkMain(dataset_train.get_n_features(),hyperparameter["channels"],num_classes= dataset_train.get_n_classes() , 
                      layers = hyperparameter["layers"], auxiliary = False,drop_prob = hyperparameter["p"], genotype = gen, binary = binary)
  model = model.cuda(device = cuda_device)
  """
  ### Train the model
  """
  train_model(model , hyperparameter, trainloader , hyperparameter["epochs"], batch_size , cuda_device, graph = plot_queue, binary = binary) 

  model.eval()
  dataset_test = Test_TEPS(binary = False,samples_per_class = 500, one_hot = False)
  testloader = torch.utils.data.DataLoader(
                  dataset_test,collate_fn = collate_fn_padd,shuffle = True,
                  batch_size=batch_size,drop_last = True)




if __name__ == "__main__":
  import csv
  
  hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 1, 'normal_index_1_0': 2, 'normal_index_1_1': 1, 'normal_index_2_0': 3, 'normal_index_2_1': 2, 'normal_index_3_0': 2, 'normal_index_3_1': 4, 'normal_node_0_0': 'dil_conv_3x3', 'normal_node_0_1': 'none', 'normal_node_1_0': 'sep_conv_5x5', 'normal_node_1_1': 'max_pool_3x3', 'normal_node_2_0': 'avg_pool_3x3', 'normal_node_2_1': 'sep_conv_7x7', 'normal_node_3_0': 'skip_connect', 'normal_node_3_1': 'sep_conv_7x7', 'reduction_index_0_0': 0, 'reduction_index_0_1': 0, 'reduction_index_1_0': 1, 'reduction_index_1_1': 1, 'reduction_index_2_0': 1, 'reduction_index_2_1': 3, 'reduction_index_3_0': 2, 'reduction_index_3_1': 2, 'reduction_node_0_0': 'none', 'reduction_node_0_1': 'skip_connect', 'reduction_node_1_0': 'sep_conv_7x7', 'reduction_node_1_1': 'sep_conv_5x5', 'reduction_node_2_0': 'dil_conv_5x5', 'reduction_node_2_1': 'skip_connect', 'reduction_node_3_0': 'skip_connect', 'reduction_node_3_1': 'sep_conv_5x5'}

  df = pd.DataFrame(columns = ["accuracy", "recall", "confusion_matrix","confusions_matrix_all","tpr","fpr","thresholds","auc"])
  for i in range(20):
      a,r,cm,cm_all,tpr,fpr,t,auc = _compute(hyperparameter, binary = False)
      df.loc[len(df.index)] = [a,r, cm,cm_all,tpr,fpr,t,auc]


  




import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import sys
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)


def _compute(hyperparameter,cuda_device, JSON_CONFIG ):
  ### Configuration 
  with open(JSON_CONFIG) as f:
    data = json.load(f)
    SETTINGS = data["WORKER_CONFIG"]
    SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  acc = []
  recall = []
  if cuda_device == None:
     cuda_device = 3
  torch.cuda.empty_cache()
  #print(hyperparameter)
  torch.cuda.set_device(cuda_device)
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:
  ##Dataset Initialisation
  name = SETTINGS["DATASET_CONFIG"]["NAME"]
  if "AUGMENTATIONS" in SETTINGS:
    augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
  else: 
    augs = None
  if SETTINGS["RESAMPLES"]:
    dataset = UEA_Full(name, cuda_device,augmentation = augs)
    kfold = KFold(n_splits = 2, shuffle = True)
    splits = [(None,None)]*SETTINGS["RESAMPLES"]
    train_dataset = dataset
    test_dataset = dataset
  elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
    train_dataset = UEA_Train(name,cuda_device,augmentation = augs )
    test_dataset = UEA_Test(name,cuda_device)
    splits = [(None,None)]
  elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
    dataset = UEA_Full(name, cuda_device)
    kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = True)
    splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())
    train_dataset = dataset
    test_dataset = dataset

  n_classes = train_dataset.get_n_classes()
  multibatch = False
  torch.cuda.empty_cache()
  
  
  for fold, (train_ids, test_ids) in enumerate(splits):    
    print('---Fold No.--{}--------------------'.format(fold))
    torch.cuda.empty_cache()
    if SETTINGS["RESAMPLES"]:
      train_ids, test_ids = next(kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy()))
    if SETTINGS["CROSS_VALIDATION_FOLDS"]: 
      # Sample elements randomly from a given list of ids, no replacement.
      train_subsampler = torch.utils.data.SubsetRandomSampler(train_ids)
      test_subsampler = torch.utils.data.SubsetRandomSampler(test_ids)
      trainloader = torch.utils.data.DataLoader(
                              dataset,collate_fn = collate_fn_padd,sampler = train_subsampler,
                              batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd,sampler = test_subsampler,
                          batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
    else:
      trainloader = torch.utils.data.DataLoader(
                              train_dataset,collate_fn = collate_fn_padd,shuffle = True,
                              batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                          batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)

    n_classes = test_dataset.get_n_classes()
    evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
    print("classes: {} - name: {}".format(train_dataset.get_n_classes(),name))
    #g = GraphConfigSpace(50)
    #s = g.sample_configuration()
    #s = s[0]
    model = ModelGraph(train_dataset.get_n_features(),64,train_dataset.get_n_classes(),train_dataset.x.shape[2],hyperparameter["graph"],hyperparameter["ops"],device = cuda_device)
    if SETTINGS["COMPILE"]:
      torch.set_float32_matmul_precision('high')
      model = torch.compile(model)
    model = model.cuda(device = cuda_device)
    """
    ### Train the model
    """
    params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("Size: {}".format(params))
    train_model(model , SETTINGS, trainloader , cuda_device,logger = False, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None) 
  #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
    torch.cuda.empty_cache()
    model.eval()
    evaluator.forward_pass(model, testloader,SETTINGS["BINARY"])
    evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"])
    total = evaluator.T()
    acc.append( evaluator.T_ACC())
    recall.append(evaluator.TPR(1))
    recall_total = evaluator.P(1)
    print("Accuracy: ", "%.4f" % ((acc[-1])*100), "%")
    print("Recall: ", "%.4f" % ((recall[-1])*100), "%")
    torch.save(model.state_dict(),"{}/weights/{:.02f}-{}".format(SAVE_PATH,acc[-1],hyperparameter["ID"]))
  acc_ = np.mean(acc)
  recall_ = np.mean(recall)
  print("Average Accuracy: ", "%.4f" % ((acc_)*100), "%")
  return acc_, recall_,params


if __name__ == "__main__":
    with open(sys.argv[1]) as f:
      HP = json.load(f)["WORKER_CONFIG"]
      HP["ID"] = "val"
    _compute(HP,3,sys.argv[1])



import csv
import networkx as nx
import json
import math
import os 
import copy
import torch
from multiprocessing import Pool
from HPO.utils.graph_utils import gen_iter
import time
import collections
import random
import matplotlib.pyplot as plt 
import numpy as np
from ConfigSpace.configuration_space import Configuration
import ConfigSpace.util as csu
from HPO.utils.visualisation import plot_scores
from HPO.algorithms.algorithm_utils import train_eval 
###Tounament Selection/Aging Selection
  #
  #parameters:
  # S : (int) number of solutions to sample at each iteration
  # P : population size
  # C : Total cycles of evolution 
  # parent : highest-accuracy model in sample
  # history : Record of all models
  # population : Currently active models


# IDEAS
#
# - Run aging evolution for "species" with 1 species per core allowing cross-over occationally
#g
#
#
#
#
#

def F1(acc,rec):
    b = 1
    if acc != None and rec != None:
      acc += 0.0001
      rec += 0.0001
      
    return (1+b**2)+ ((acc*rec)/(((b**2)*(acc**-1))+(rec**-1)))

class Model:
  MUTATE_RATE = 0.2
  def __init__(self, cs,ops_list):
    self.graph = None
    self.op_list = ops_list
    self.ops = None 
    self.accuracy = None
    self.cs = cs
    self.recall = None

  def set_arch(self, arch):
    self._arch = arch
    self.graph = arch["graph"]
    self.ops = arch["ops"]
  def exchange_one( self , key  ):
    n_idx = []
    for idx ,i in enumerate(key):
      if i.isdigit():
        n_idx.append(idx)
    new_key = key[:n_idx[0]]+"1"+key[n_idx[-1]+1:]
    while True:
      new = self.cs.sample_configuration()[0]["ops"]
      if new[new_key] != self.ops[key]:
        print("Mutated {} to {}".format(self.ops[key],new[new_key]))
        self.ops[key] = new[new_key]
        return
  def generate_op(self):
    new = self.cs.sample_configuration()[0]["ops"]
    idx = len(self.graph) -1
    if "op_{}".format(idx) in list(self.ops.keys()):
      print("op exists!")
      print(self.graph)
      print(self.ops)
      exit()
    else:
      self.ops["op_{}".format(idx)] = new["op_1"]    
      self.ops["op_{}_dil".format(idx)] = new["op_1_dil"]    
      self.ops["op_{}_kernel".format(idx)] = new["op_1_kernel"]    
      self.ops["op_{}_stride".format(idx)] = new["op_1_stride"]
      print("added {} as operation {} for edge {}".format(new["op_1"],idx,self.graph[-1]))
  def __repr__(self):
    return "Model(Accuracy: {} , Recall: {})".format(self.accuracy,self.recall)

  def set_value(self, name, value):
    self._arch[name] = value
  def get_params(self):
    return list(self.arch().keys())
  def arch(self):
    return self._arch

def train_and_eval_population(worker, population, SETTINGS, train):
  configs = []
  for model in population:
    if model.accuracy == None:
      configs.append({"graph":model.graph,"ops":model.ops})
  acc , _rec , _config = train.eval( configs )
  for mod,result,recall in zip(population, acc, _rec):
    mod.accuracy = result
    mod.recall = recall

def model_change(parent, child):
  p_dict , c_dict = parent.arch().get_dictionary(), child.arch().get_dictionary()
  for _,i in zip(p_dict, c_dict):
    if p_dict[i] != c_dict[i]:
      print("Mutated {} from {} to {}".format(i, p_dict[i], c_dict[i]))
      return True
  print("WARNING: no change found!!!")
  return False



def Mutate(cs, parent_model : Model) -> Model:
  model = copy.deepcopy(parent_model) 
  model.accuracy = None
    
  def op_mutation(model: Model) -> Model:
    operation = random.choice(list(model.ops.keys()))
    model.exchange_one(operation)
    return model
     
  def graph_mutation(model : Model) -> Model:
    old = copy.deepcopy(model.graph)
    model.graph = gen_iter(model.graph,enable_new_sources = False,rate = 0.5)
    if len(old) == len(model.graph):
      print("graph mutate failed")
      print(old)
      print(model.graph)
      exit()
    model.generate_op()
    for i in model.graph:
      if not i in old:
        print("Mutated adding edge {}".format(i))
        return model
    print("No difference in graphs")
    print(old)
    print(model.graph)
    exit()
  
  return random.choice([op_mutation, graph_mutation])(model)    

def regularized_evolution(configspace, worker ,SEARCH_SETTINGS : dict,json_config):
  population = []
  history = []  # Not used by the algorithm, only used to report results.

  CS = configspace
  # Initialize the population with random models.
  train = train_eval( worker , json_config)

  while len(population) < SEARCH_SETTINGS["POPULATION_SIZE"]:
    model = Model( CS ,ops_list = SEARCH_SETTINGS["OPS"])
    model.set_arch( CS.sample_configuration()[0] )
    population.append(model)
    history.append(model)
  train_and_eval_population(worker, population, SEARCH_SETTINGS, train )
  # Carry out evolution in cycles. Each cycle produces a model and removes
  # another.
  children = []
  print_counter = 0
  while len(history) < SEARCH_SETTINGS["TOTAL_EVALUATIONS"]:

    print("Starting cycle", len(history) - SEARCH_SETTINGS["POPULATION_SIZE"])
    # Sample randomly chosen models from the current population.

    if len(children) < SEARCH_SETTINGS["CORES"]:
      sample = []
      while len(sample) < SEARCH_SETTINGS["SAMPLE_SIZE"]:
        # Inefficient, but written this way for clarity. In the case of neural
        # nets, the efficiency of this line is irrelevant because training neural
        # nets is the rate-determining step.
        candidate = random.choice(list(population))
        sample.append(candidate)

      # The parent is the best model in the sample.
      print("Selecting Parent via F1 Score")
      parent = max(sample, key=lambda i: i.accuracy)
      print("Scores of Parent F1: {} -- Acc: {} -- Recall: {}".format(F1(parent.accuracy,parent.recall),parent.accuracy,parent.recall))
      #parent = max(sample, key=lambda i: i.accuracy)

      # Create the child model and store it.
      print("Selected Parent: {}".format(parent.graph))
      child = Mutate(CS,parent)
      print("Mutated child: {}".format(child.graph))
      children.append(child)

    else:
      train_and_eval_population(worker, children, SEARCH_SETTINGS["CORES"], train)
      for i in children:
        population.append(i)
        history.append(i)
        population.pop(0)
      children = []
      # Remove the oldest model.
      best = max(population, key=lambda i: i.accuracy)
      print("--Current best model--")
      print("Accuracy:{} -- Recall: {} ".format(best.accuracy,best.recall))
      print("Architecture: ", best.arch())
      print("Population Size: ", len(population))
      print("Total Evaluations: ", len(history))
      g = nx.DiGraph()
      g.add_edges_from(best.graph)
      #print("nodes in topological sort:{}".format(len([x for x in nx.topological_sort(g)])))
      fixed = {"S":(-10000,0),"T":(20,0)}
      node = fixed.keys()
      pos = nx.spring_layout(g)#,k=0.5, iterations=200
      plt.figure(figsize = (19,12))
      e_labels = {}
      for i,(e,o) in enumerate(zip(best.graph,best.ops)):
        e_labels[e] = "{} (k:{},s:{},d:{})".format(best.ops["op_{}".format(i)],best.ops["op_{}_kernel".format(i)],best.ops["op_{}_stride".format(i)],best.ops["op_{}_dil".format(i)])
      nx.draw(
        g, pos, edge_color='black', width=1, linewidths=1,
        node_size=500, node_color='pink', alpha=0.9,
        labels={node: node for node in g.nodes()}
        )
      nx.draw_networkx_edge_labels(
      g, pos,
      edge_labels=e_labels,
      font_color='red')
      plt.axis('off')
      plt.savefig("{}/best_mode_iter_{}".format(SEARCH_SETTINGS["PATH"],len(history)))
      print_counter +=1
      print_counter = 0
      accuracy_scores = []
      for i in history:
        accuracy_scores.append(i.accuracy)

      plot_scores(accuracy_scores)
  return history


def main(worker, configspace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]

  history = regularized_evolution(configspace, worker, SETTINGS,json_config)

if __name__ == '__main__':
  main()





def nats_encode(config): 
  paths = [[""],["",""],["","",""]]
  for i in config:
    paths[int(i[-1])-1][int(i[0])] =  config[i]+"~"+str(int(i[0]))
  for i in paths:
    paths[paths.index(i)] = "|"+"|".join(i)+"|"
  return "+".join(paths)
